#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import functools
import logging
import os
import time
import urllib.error

import AUR.PkgList
import AUR.RPC
import AUR.SRCINFO
import XCGF
import XCPF.RepoWrapper



################################## Functions ###################################

def download_git_repo_by_pkgbase_wrapper(outdir, pkgbase):
  '''
  Wrapper around AUR.RPC.download_git_repo_by_pkgbase for download jobs.
  '''
  AUR.RPC.download_git_repo_by_pkgbase(outdir, pkgbase)
  return pkgbase



############################## Buildable Packages ##############################

class AurPkg(XCPF.Pkg.Pkg):
  # TODO
  # Remove the .SRCINFO retrieval once the RPC interface catches up.
  def __init__(self, arch, pkg, asi=None):
    '''
    arch:
      The target architecture.

    pkg:
      An Aur package from the AUR.RPC module. It's just a dictionary with the
      info from the AUR's JSON interface.

    asi:
      An AUR.SRCINFO.AurSrcinfo object.
    '''
    # Invoke the parent class's init method with True because AUR packages must
    # be built.
    super(self.__class__, self).__init__(True)
    self.arch = arch
    self.pkg = pkg
    self.pkginfo = None
    if asi is None:
      asi = AUR.SRCINFO.AurSrcinfo()
    self.asi = asi


  # Lazy retrieval.
  def get_pkginfo(self):
    '''
    Internal method to get package information from retrieved SRCINFO. This will
    defer retrieval until it's necessary.
    '''
    if self.pkginfo is None:
      try:
        pkgbases_and_pkgnames = ((self.pkg['PackageBase'], self.pkg['Name']),)
        for si in self.asi.get_pkginfo(pkgbases_and_pkgnames, arch=self.arch):
          self.pkginfo = si
          break
      except urllib.error.HTTPError as e:
        raise XCPF.Pkg.PkgError(
          'failed to retrieve .SRCINFO for {}'.format(self.pkg.name),
          error=e
        )
    return self.pkginfo


  def display_fields(self, extended=False):
    '''
    See XCPF.Pkg.Pkg.display_fields
    '''
    return AUR.RPC.display_fields(self.pkg, extended=extended)

  def maintainers(self):
    '''
    See XCPF.Pkg.Pkg.maintainers
    '''
    return self.pkg['Maintainer']

  def rid(self):
    '''
    See XCPF.Pkg.Pkg.rid
    '''
    return time.strftime(
      XCGF.DISPLAY_TIME_FORMAT,
      time.gmtime(self.pkg['LastModified'])
    )

  def pkgname(self):
    '''
    See XCPF.Pkg.Pkg.pkgname
    '''
    return self.pkg['Name']

  def version(self):
    '''
    See XCPF.Pkg.Pkg.version
    '''
    return self.pkg['Version']

  def pkgdesc(self):
    '''
    See XCPF.Pkg.Pkg.pkgdesc
    '''
    return self.pkg['Description']

  def pkgbase(self):
    '''
    See XCPF.Pkg.Pkg.pkgbase
    '''
    return self.pkg['PackageBase']

  def repo(self):
    '''
    See XCPF.Pkg.Pkg.repo
    '''
    return 'AUR'

  def depends(self):
    '''
    See XCPF.Pkg.Pkg.depends
    '''
    return self.get_pkginfo().depends

  def makedepends(self):
    '''
    See XCPF.Pkg.Pkg.makedepends
    '''
    return self.get_pkginfo().makedepends

  def checkdepends(self):
    '''
    See XCPF.Pkg.Pkg.checkdepends
    '''
    return self.get_pkginfo().checkdepends

  def provides(self):
    '''
    See XCPF.Pkg.Pkg.provides
    '''
    return self.get_pkginfo().provides

  def conflicts(self):
    '''
    See XCPF.Pkg.Pkg.conflicts
    '''
    return self.get_pkginfo().conflicts

  def replaces(self):
    '''
    See XCPF.Pkg.Pkg.replaces
    '''
    return self.get_pkginfo().replaces

  def groups(self):
    '''
    See XCPF.Pkg.Pkg.groups
    '''
    try:
      return self.pkg['Groups']
    except KeyError:
      return iter([])



############################ AurRepoWrapperFactory #############################

class AurRepoWrapperFactory(XCPF.RepoWrapper.RepoWrapperFactory):
  '''
  RepoWrapperFactory for the AUR.
  '''
  def wrap(self, env, reponame):
    if reponame != 'AUR':
      raise ValueError('{} only supports the AUR repo'.format(self))

    try:
      aur = env['aur']
    except KeyError:
      aur = AUR.RPC.AurRpc(ttl=env[XCPF.RepoWrapper.ENV_KEY_TTL])
      env['aur'] = aur

    try:
      asi = env['asi']
    except KeyError:
      asi = AUR.SRCINFO.AurSrcinfo(ttl=env[XCPF.RepoWrapper.ENV_KEY_TTL])
      env['asi'] = asi

    try:
      apl = env['apl']
    except KeyError:
      apl = AUR.PkgList.PkgList(ttl=env[XCPF.RepoWrapper.ENV_KEY_TTL], auto_refresh=True)
      env['apl'] = apl

    return env, AurRepoWrapper(env)



############################# OfficialRepoWrapper ##############################

class AurRepoWrapper(XCPF.RepoWrapper.RepoWrapper):
  '''
  RepoWrapper for the AUR.
  '''
  def __init__(self, env):
    '''
    Initialize the object. The complexity of wrapping the AUR is nested in the
    AurRpc, AurSrcinfo and PkgList objects. These will be configured in the
    common environment object by the RepoWrapperFactory.
    '''
    # Invoke the base class's __init__ method as required with the environment
    # object and the repo's name.
    super(AurRepoWrapper, self).__init__(env, 'AUR')


  def aur_to_buildable_pkg(self, pkg):
    '''
    Internal method used by other methods of this class to convert an AUR.RPC
    package to AurPkg, which is a subclass of XCPF.Pkg.Pkg.
    '''
    arch = self.env[XCPF.RepoWrapper.ENV_KEY_CPO].arch()
    return AurPkg(arch, pkg, asi=self.env['asi'])


  def refresh(self, force=False):
    '''
    Clear the cached data.
    '''
    self.env['aur'].refresh(force=force)
    self.env['apl'].refresh(force=force)
    yield from ()


  def satisfy(self, targets, check_groups=True, replacements=True):
    '''
    Search the AUR for targets. Currently both "groups" and "replacements" are
    ignored due to limitations in the RPC interface.
    '''
    # Retrieving metadata for all AUR packages is unreasonable so return
    # nothing as specified by XCPF.RepoWrapper.RepoWrapper
    if not targets:
      return XCPF.Pkg.PkgSet(), dict(), set()

    satisfied = list()
    unsatisfied = set(targets)

    ver_reqs = XCPF.collect_version_requirements(targets)

    maybe_aur = dict(
      (name, repo) for repo, name in (XCPF.repo_and_package(f) for f in targets)
      if repo is None or repo == 'AUR'
    )
    satisfied.extend(
      self.env['aur'].info(XCPF.common.get_required_version_strings(ver_reqs, maybe_aur))
    )
    unsatisfied -= set(
      '{}/{}'.format(repo, name) if repo else name for name, repo in (
        (a, maybe_aur[a]) for a in (p['Name'] for p in satisfied)
      )
    )

    satisfied = XCPF.Pkg.PkgSet(
      self.aur_to_buildable_pkg(p) for p in satisfied
    )
    return satisfied, dict(), unsatisfied


  def search(self, terms, by=XCPF.RepoWrapper.SEARCH_BYS[0], intersect=False):
    '''
    Search the AUR via the RPC interface.
    '''
    for p in sorted(
      self.env['aur'].search(terms, by=by, intersect=intersect), key=lambda x: x['Name']
    ):
      yield self.aur_to_buildable_pkg(p)


  def list(self):
    '''
    List all packages in the AUR. Only the name is listed because the
    packagelist does not contain versions and descriptions.
    '''
    for p in sorted(self.env['apl']):
      yield p, None, None


  def download_pkgbuilds(self, outdir, pkgbases, with_source_files=False):
    '''
    Download PKGBUILD git repos.
    '''
    pkgbase_jobs = dict()

    for pkgbase in pkgbases:
      pkgbase_jobs[pkgbase] = [
        XCGF.AsyncJob(
          download_git_repo_by_pkgbase_wrapper,
          args=(outdir, pkgbase)
        )
      ]

    if with_source_files:
      pkgbase_jobs = XCPF.RepoWrapper.append_source_download_jobs(
        self.env,
        outdir,
        self.env['asi'].get_srcinfo(pkgbases),
        pkgbase_jobs
      )

    def callback(pkgbase, results):
      if all(results[0]):
        git_url = AUR.common.AUR_GIT_URL_FORMAT.format(
          urllib.parse.quote_plus(pkgbase)
        )
        qualified_pkgbase = '{}/{}'.format(self.name, pkgbase)
        logging.info('found {} at {}'.format(qualified_pkgbase, git_url))
        yield qualified_pkgbase

    grouped_jobs = list(
      {
        'callback': functools.partial(callback, pkgbase),
        'jobs': jobs
      } for pkgbase, jobs in pkgbase_jobs.items()
    )

    return set(), grouped_jobs

