#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2015  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import argparse
import AUR.common
import json
import logging
import MemoizeDB
import os.path
import re
import sqlite3
import sys
import urllib.parse
import XCGF
import xdg.BaseDirectory
import XCPF.SRCINFO



################################### Globals ####################################

SRCINFO_PATH = '/cgit/aur.git/plain/.SRCINFO?h={}'
PKGLIST_URL = AUR.common.AUR_URL + SRCINFO_PATH



################################## Functions ###################################

def srcinfo_dbpath():
  '''
  Return a path to the .SRCINFO caching database.
  '''
  cachedir = xdg.BaseDirectory.save_cache_path(AUR.common.XDG_NAME)
  return os.path.join(cachedir, 'SRCINFO.sqlite3')



def srcinfo_url(pkgname):
  '''
  Get the URL of the .SRCINFO file.
  '''
  return PKGLIST_URL.format(urllib.parse.quote_plus(pkgname))



################################## AurSrcInfo ##################################

class AurSrcinfo(object):
  '''
  A caching AUR .SRCINFO retriever.
  '''
  SRCINFO_TABLE = 'srcinfo'

  def __init__(self, mdb=None, dbpath=None, ttl=AUR.common.DEFAULT_TTL):
    if mdb is None:
      if dbpath is None:
        dbpath = srcinfo_dbpath()
      def f(pkgnames):
        for pkgname in pkgnames:
          url = srcinfo_url(pkgname)
          try:
            yield pkgname, (XCGF.text_from_url(url),)
          except urllib.error.HTTPError as e:
            if e.code == 404:
              yield pkgname, None
            else:
              raise e
      glue = {
        self.SRCINFO_TABLE : (f, (('text', 'TEXT'),), ttl)
      }
      conn = sqlite3.connect(
        dbpath,
        detect_types=(sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
      )
      mdb = MemoizeDB.MemoizeDB(conn, glue)
      mdb.db_initialize()
    self.mdb = mdb


  def get(self, pkgbases):
    '''
    Iterate over .SRCINFO data. Returns None if the file could not be retrieved.
    This expects pkgbase arguments.
    '''
    for srcinfo in self.mdb.get_nth_field_many(self.SRCINFO_TABLE, pkgbases):
      yield srcinfo


  def get_srcinfo(self, pkgbases):
    '''
    Retrieve parsed .SRCINFO.

    This returns instances of XCPF.SRCINFO.Srcinfo.
    '''
    for srcinfo_text in self.get(pkgbases):
      yield XCPF.SRCINFO.Srcinfo(text=srcinfo_text)


  def get_pkginfo(self, pkgbases_and_pkgnames, arch=None):
    '''
    Retrieve package information.

    pkgbases_and_pkgnames:
      An iterator over package base and package name pairs.

    arch:
      The target architecture.
    '''
    pkgbases, pkgnames = zip(*pkgbases_and_pkgnames)
    srcinfo = dict()
    for pkgbase, text in zip(pkgbases, self.get(pkgbases)):
      if text:
        srcinfo[pkgbase] = XCPF.SRCINFO.Srcinfo(text=text)
    for pkgbase, pkgname in zip(pkgbases, pkgnames):
      try:
        info = srcinfo[pkgbase]
      except KeyError:
        yield None
      else:
        yield info.pkginfo(pkgname, arch=arch)



##################################### Main #####################################

def main(args=None):
  argparser = argparse.ArgumentParser(
    description='Retrieve AUR .SRCINFO files and print them to STDOUT.'
  )
  argparser.add_argument(
    'pkgbase', nargs='+'
  )
  pargs = argparser.parse_args(args)
  a = AurSrcinfo()
  for s in a.get(pargs.pkgbase):
    print(s)


if __name__ == '__main__':
  try:
    main()
  except KeyboardInterrupt:
    pass
