#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2015 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import calendar
import collections
import hashlib
import io
import itertools
import json
import logging
import multiprocessing.dummy as mp
import os
import pwd
import shlex
import shutil
import sys
import time
import types
import urllib.request
import warnings

################################## Constants ###################################

DISPLAY_TIME_FORMAT = '%Y-%m-%d %H:%M:%S %Z'
RFC_2822_TIME_FORMAT = '%a, %d %b %Y %H:%M:%S %Z'

SIG_EXT = '.sig'



##################################### HTTP #####################################

def text_from_url(url):
  '''
  Retrieve text from a URL.
  '''
  logging.debug('retrieving data from {}'.format(url))
  with urllib.request.urlopen(url) as f:
    return f.read().decode()



##################################### JSON #####################################

def load_json(path):
  '''
  Simple wrapper function to open a file and load the JSON in it.
  '''
  with open(path, 'r') as f:
    return json.load(f)



def load_json_from_url(url):
  '''
  Simple wrapper to load a JSON object from a remote URL.
  '''
  logging.debug('retrieving JSON data from {}'.format(url))
  with urllib.request.urlopen(url) as f:
    return json.loads(f.read().decode())


class InformationJSONEncoder(json.JSONEncoder):
  '''
  JSON Encoder for pretty-printing arbitrary objects. Non-serializable objects
  will be replaced with their string representation via repr().
  '''
  def default(self, obj):
    return repr(obj)



################################### Network ####################################

def mirror(url, path, cache_time=-1):
  '''
  Download a file at the given URL to a local path. The 'If-Modified-Since'
  header is used to avoid unnecessary downloads when supported by the server. A
  further option is provided to cache local files for a given interval before
  updating them.

  If the local file does not exist, the remote file will always be downloaded.

  url:
    The remote URL.

  path:
    The local path.

  cache_time:
    If negative, the file will be downloaded only if the server
    reports that a newer version is available by comparing the "Last-Modified"
    to the local file's last modification time.

    If 0, the file will be downloaded regardless of the last modification times.

    If positive, the file will only be downloaded if the local file's last
    modification time is older than this number of seconds AND the server has
    a newer version as determined by the "If-Modified-Since" header.
  '''
  headers = dict()

  if cache_time is None:
    cache_time = -1

  if cache_time != 0:
    try:
      mtime = os.path.getmtime(path)
    except FileNotFoundError:
      pass
    else:
      headers['If-Modified-Since'] = time.strftime(RFC_2822_TIME_FORMAT, time.gmtime(mtime))
      # The "> 0" check should be unnecessary but the user could mess with the mtime.
      if cache_time > 0 and (time.time() - mtime) < cache_time:
          return

  req = urllib.request.Request(url, headers=headers)
  logging.debug('requesting {}'.format(url))
  with urllib.request.urlopen(req) as f:
    # The file has not been modified.
    if f.code == 304:
      logging.debug('unmodified')
      return
    else:
      logging.debug('saving {} to {}'.format(url, path))
      # Download the file.
      with open(path, 'wb') as g:
        buf = f.read(io.DEFAULT_BUFFER_SIZE)
        g.write(buf)
        while(buf):
          buf = f.read(io.DEFAULT_BUFFER_SIZE)
          g.write(buf)
      # Synchronize the modification time with the server.
      try:
        server_mtime = f.headers['Last-Modified']
      except KeyError:
        pass
      else:
        if server_mtime is not None:
          mtime = calendar.timegm(time.strptime(server_mtime, RFC_2822_TIME_FORMAT))
          os.utime(path, (mtime, mtime))



def uri_parse_path_parameters(path):
  '''
  Parse URI path parameters. See e.g.

      https://tools.ietf.org/html/rfc3986#page-23

  path:
    The path component of a URI as would be returned by urllib.parse.urlsplit().

  Each path component is yielded along with a tuple of name-value pairs where
  the value is None if only the name was provided.
  '''
  for component in path.split('/'):
    try:
      component, parameters = component.split(';', 1)
    except ValueError:
      yield component, None
      continue

    params = tuple(
      (tuple(p.split('=', 1)) if '=' in p else (p, None))
      for p in parameters.split(',')
    )



def uri_strip_path_parameters(path):
  '''
  Strip path paramters from a URI path. See uri_parse_path_parameters().

  Returns the path without the parameters.
  '''
  return '/'.join(x[0] for x in uri_parse_path_parameters(path))



#################################### Pushd #####################################

class Pushd(object):
  '''
  Temporarily change the working directory.
  '''
  def __init__(self, path=None):
    if path is None:
      self.path = os.getcwd()
    else:
      self.path = os.path.abspath(path)
    self.orig = None

  def __enter__(self):
    self.orig = os.getcwd()
    os.chdir(self.path)
    return self.path

  def __exit__(self, typ, val, traceback):
    os.chdir(self.orig)



################################## Lock File ###################################

class LockError(Exception):
  '''
  Errors raised by the lock file class.
  '''
  def __init__(self, path, msg, error=None):
    self.path = path
    self.msg = msg
    self.error = error

  def __str__(self):
    s = '{}({}): {}'.format(self.__class__.__name__, self.path, self.msg)
    if self.error:
      s += ' [{}]'.format(self.error)
    return s



class Lockfile(object):
  '''
  Lock file class.
  '''
  def __init__(self, path, name, mode=0o644):
    '''
    path:
      The path to the lock file. If None, this will do nothing.

    name:
      The resource name.

    mode:
      The mode of the lock file.
    '''
    if path:
      path = os.path.abspath(path)
    self.path = path
    self.locked = False
    self.name = name
    self.mode = mode

  def error(self, locking, error=None):
    if locking:
      prefix = ''
    else:
      prefix = 'un'
    msg = 'failed to {}lock {} [{}]'.format(prefix, self.path, self.name)
    le = LockError(self.path, msg, error=error)
    logging.debug(str(le))
    raise le

  def __enter__(self):
    if not self.path:
      return None
    try:
      # Use O_EXCL to ensure that this will only succeed if no previous file
      # exists.
      flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
      fd = os.open(self.path, flags, mode=self.mode)
      with os.fdopen(fd, 'w') as f:
        f.write(str(os.getpid()))
    except OSError as e:
      self.error(True, error=e)
    else:
      self.locked = True
      return self.path

  def __exit__(self, typ, val, traceback):
    if self.locked:
      try:
        os.unlink(self.path)
      except OSError as e:
        self.error(False, error=e)
      else:
        self.locked = False



################################ File Functions ################################

def get_checksum(path, typ):
  '''
  Return a checksum of a local file as a hexadecimal digest.
  '''
  h = hashlib.new(typ)
  b = h.block_size
  try:
    with open(path, 'rb') as f:
      buf = f.read(b)
      while buf:
        h.update(buf)
        buf = f.read(b)
    return h.hexdigest()
  except FileNotFoundError:
    return None



def copy_file_and_maybe_sig(from_path, to_path, sig=False):
  '''
  Copy a file and its signature, if it exists.
  '''
  shutil.copyfile(from_path, to_path)
  if sig:
    a = from_path + SIG_EXT
    b = to_path + SIG_EXT
    try:
      shutil.copyfile(a, b)
    except FileNotFoundError as e:
      logging.debug('no matching sig {}'.format(from_path))
      return False
    else:
      return True
  else:
    return False



def relink(target, path):
  '''
  Ensure that path is a symlink that points to the target.
  '''
  try:
    if os.readlink(path) == target:
      return
    else:
      os.remove(path)
  except FileNotFoundError:
    pass
  except OSError as e:
    if e.errno != errno.EINVAL:
      raise e
    else:
      os.remove(path)
  os.symlink(target, path)



def copy_if_absent_or_newer(from_path, to_path):
  '''
  Copy a file if the destination is absent or the source is newer.
  '''
  from_mtime = os.path.getmtime(from_path)
  os.makedirs(os.path.dirname(to_path), exist_ok=True)
  try:
    to_mtime = os.path.getmtime(to_path)
  except FileNotFoundError:
    pass
  else:
    if from_mtime <= to_mtime:
      return
  shutil.copy(from_path, to_path)
  os.utime(to_path, (from_mtime, from_mtime))



############################### Argument Parsing ###############################

def expand_short_args(args):
  '''
  Expand short arguments.

  Example: ["-Syu"] -> ["-S", "-y", "-u"]

  Arguments following '--' will not be expanded.
  '''
  expand = True
  for arg in args:
    if expand:
      try:
        if arg[0] == '-':
          if arg[1] == '-':
            expand = bool(arg[2:])
            yield arg
          else:
            for c in arg[1:]:
              yield '-' + c
          continue
      except IndexError:
        pass
    yield arg




def filter_arguments(args, remove):
  '''
  Iterate over arguments while omitting some.

  remove:
    A dictionary of the arguments to recognize and the number of parameters to
    discard. If the value is an interger then that number of arguments will be
    discarded after the recognized argument. If the argument is a string then
    arguments will be discarded until the next one that begins with that string.
  '''
  itr = iter(args)
  prefix = None
  for a in itr:
    if prefix is not None:
      if not a.startswith(prefix):
        continue
      else:
        prefix = None

    try:
      v = remove[a]
    except KeyError:
      yield a
    else:
      if isinstance(v, int):
        for i in range(v):
          a = next(itr)
      elif isinstance(v, str):
        prefix = v
      else:
        raise ValueError('filter_arguments: values in "remove" argument must be of type int or str, not {}'.format(type(v)))




################################ Nested Values #################################

def recursive_get(obj, keys, default=None):
  '''
  Recursive version of built-in get() function.

  Attempt to retrieve a value from an object using a list of nested keys. Return
  the default value if any KeyError occurs. This is useful for JSON objects with
  multiple nested mappings that may not exist.

  keys:
    The keys to the nested value. For example, to reach foo['bar']['baz'],
    ks should be ('bar', 'baz').
  '''
  target = obj
  for k in keys:
    try:
      target = target[k]
    except KeyError:
      return default
  return target



def recursive_getattr(obj, attrs, default=None):
  '''
  Recursive version of built-in get() function.

  Attempt to retrieve a value from an object using a list of nested attributes.
  Return the default value if any AttributeError occurs.

  attrs:
    The attributes to the nested value. For example, to reach foo.bar.baz,
    attrs should be ('bar', 'baz').
  '''
  target = obj
  for a in attrs:
    try:
      target = getattr(target, a)
    except AttributeError:
      return default
  return target



##################################### Bash #####################################

def sh_comment_header(title):
  '''
  Section header comment.
  '''
  return '{:#^80s}'.format('# {} #'.format(title))



def sh_prompt_continue(question, expected_char):
  '''
  Prompt the user with a question and exit if the expected character is not
  given.
  '''

  c = expected_char[0]
  ans='{}{}'.format(c.upper(), c.lower())

  return '''{header}
read -p {question} -n 1 -r ans
echo
if [[ ! $ans =~ ^[{ans}]$ ]]
then
  exit 0
fi
'''.format(
    header=sh_comment_header('Prompt user before continuing.'),
    question=shlex.quote(question + ' '),
    ans=ans
  )



def interpolate_words(words, kwargs=None):
  '''
  Insert keyword arguments into given words using the string format method.
  '''
  if kwargs:
    for w in words:
      if isinstance(w, str):
        yield w.format(**kwargs)
      else:
        yield w
  else:
    for w in words:
      yield w


def ensure_strs(words):
  '''
  Ensure that all words in the iterable are strings.
  '''
  for w in words:
    if isinstance(w, str):
      yield w
    else:
      yield str(w)



def sh_quote_words(words, kwargs=None):
  '''
  Convert a command to a shell-escaped string. Optional interpolation is
  supported with interpolate_words if keyword arguments are given.
  '''
  if kwargs:
    words = list(interpolate_words(words, kwargs))
  return ' '.join((shlex.quote(x) for x in ensure_strs(words)))



################################### Logging ####################################

def configure_logging(
  level=None,
  info_msg_fmt='[{asctime:s}] {levelname:s}: {message:s}',
  other_msg_fmt='[{asctime:s}] {levelname:s}: {message:s} [{module:s}: {lineno:d} ({funcName:s})]',
  style='{',
  date_fmt='%Y-%m-%d %H:%M:%S',
  info_levels=(logging.INFO, logging.WARNING),
  stream=sys.stderr,
  log=None,
  color=False,
  clear_existing_handlers=False
):
  '''
  Configure logging. This will set up handlers to display additional information
  for some messages. It will also optionally configure colored output.

  level:
    Logging level defined in the logging module, e.g. logging.INFO.

  info_msg_fmt:
    The format to use for "info" messages. See "info_levels".

  other_msg_fmt:
    The format to use for non-"info" messages. See "info_levels".

  style:
    The logging.Formatter style.

  date_fmt:
    The date format.

  info_levels:
    Log levels to consider as "info" when selecting either "info_msg_fmt" or
    "other_msg_fmt" to display the message.

  stream:
    The output stream for the stream handler if "log" is not given. This should
    have an "isatty()" method when "color" is None.

  log:
    A log file to use instead of a stream. Color escapes will be omitted even
    if "color" is True.

  color:
    If True then insert ANSI color escapes in the stream output. If False then
    do not insert any color escapes. If None, call the streams isatty() method.
    to get True or False.

  clear_existing_handlers:
    If True, clear existing handlers.
  '''
  if level is None:
    level = logging.WARNING
  if color is None and not log:
    try:
      color = stream.isatty()
    except AttributeError:
      color = False

  # Inspired by
  # http://stackoverflow.com/questions/1343227/can-pythons-logging-format-be-modified-depending-on-the-message-log-level

  class LevelDependentFormatter(logging.Formatter):
    def format(self, record):
      prev_style = self._style
      if record.levelno in info_levels:
        self._style = logging._STYLES[style][0](info_msg_fmt)
      else:
        self._style = logging._STYLES[style][0](other_msg_fmt)
      output = super(LevelDependentFormatter, self).format(record)
      self._style = prev_style
      return output


  formatter_class = LevelDependentFormatter
  if color and not log:
    try:
      import colorsysplus
    except ImportError:
      pass
    else:
      level_colors = {
        'DEBUG': colorsysplus.ansi_sgr(fg='blue'),
        'INFO': colorsysplus.ansi_sgr(fg='green', bold=True),
        'WARNING': colorsysplus.ansi_sgr(fg='yellow', bold=True),
        'ERROR': colorsysplus.ansi_sgr(fg='red', bold=True),
        'CRITICAL': colorsysplus.ansi_sgr(fg='magenta', bold=True, blink=True),
      }
      time_color = colorsysplus.ansi_sgr(fg='cyan')
      reset = colorsysplus.ansi_sgr(reset=True)

      # Inspired by
      # http://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
      class ColoredFormatter(LevelDependentFormatter):
        def format(self, record):
          try:
            level_color = level_colors[record.levelname]
          except KeyError:
            pass
          else:
            record.levelname = '{}{}{}'.format(
              level_color, record.levelname, reset
            )
          return super(ColoredFormatter, self).format(record)

        def formatTime(self, *args, **kwargs):
          t = super(ColoredFormatter, self).formatTime(*args, **kwargs)
          return '{}{}{}'.format(time_color, t, reset)

      formatter_class = ColoredFormatter

  formatter = formatter_class(
    fmt=info_msg_fmt,
    datefmt=date_fmt,
    style=style
  )

  if log:
    handler = logging.FileHandler(log)
  else:
    handler = logging.StreamHandler(stream)
  handler.setFormatter(formatter)
  handlers = [handler]

  if clear_existing_handlers:
    for h in list(logging.root.handlers):
      logging.root.removeHandler(h)

  logging.root.addHandler(handler)
  logging.root.setLevel(level)

#   logging.basicConfig(
#     format=info_msg_fmt,
#     datefmt=date_fmt,
#     level=level,
#     handlers=handlers
#   )



class DisableLogger():
  '''
  Context manager to temporarily disable logging for password input.
  '''
  def __enter__(self):
    logging.disable(logging.CRITICAL)

  def __exit__(self, exc_type, exc_value, traceback):
    logging.disable(logging.NOTSET)



############################# Deprecation Warnings #############################

def warn_deprecated(msg):
  '''
  Issue deprecation warnings.
  '''
  warnings.warn(
    msg,
    category=DeprecationWarning,
  )



def deprecated(f):
  '''
  Function decorator to issue deprecation warnings.
  '''
  def new_f(*args, **kwargs):
    warn_deprecated('function {} is deprecated'.format(f.__name__))
    return f(*args, **kwargs)
  new_f.__name__ = f.__name__
  new_f.__doc__ = f.__doc__
  new_f.__dict__.update(f.__dict__)
  return new_f



################################# Environment ##################################

def get_sudo_user_info():
  '''
  Get the name, uid and gid of the sudo user, if there is one.
  '''
  usr = os.getenv('SUDO_USER')
  if usr is not None:
    usrpwd = pwd.getpwnam(usr)
    uid = usrpwd[2]
    gid = usrpwd[3]
  else:
    uid = None
    gid = None
  return usr, uid, gid



############################## Number Formatting ###############################

def decimal_offsets(ss):
  '''
  Determine the offsets relative to the decimal point of the first and last
  digits of a number represented in scientific format ("{:e}").
  '''
  for s in ss:
    s = s.lower()
    n, e = s.rsplit('e', 1)
    try:
      b,a = n.split('.',1)
      a = a.rstrip('0')
    except ValueError:
      b = n
      a = ''
    e = int(e)
    f = e - len(a)
    yield e+1, f


def human_format_for_set_of_floats(xs, pad=True):
  '''
  Return a format string that will produce strings of equal width in the
  shortest and most natural format (which I have arbitrarily chosen).
  '''
  if pad:
    pad_char = '0'
  else:
    pad_char = ''

  n = len(xs)
  p = 0
  fmt = None
  ss = None
  while True:
    fmt = '{{:0.{:d}e}}'.format(p)
    ss = list(s for s in set(fmt.format(x) for x in xs) if float(s) in xs)
    if len(ss) < n:
      p += 1
    else:
      break

  offsets = list(decimal_offsets(ss))
  max_len = max(len(s) for s in ss)
  max_off = max(o[0] for o in offsets)
  min_off = min(o[1] for o in offsets)
  max_span = max_off - min(min_off, 0)
  if min_off < 0:
    max_span += 1
    if max_off < 0:
      max_span -= max_off - 1

  if max_span <= max_len:
    if min_off >= 0:
      fmt = '{{:{}{:d}.0f}}'.format(pad_char, max_span)
    else:
      fmt = '{{:{}{:d}.{:d}f}}'.format(pad_char, max_span, abs(min_off))

  return fmt



######################### Parallelism and Concurrency ##########################

class JobError(Exception):
  '''
  Errors raised by jobs in attempt_jobs_in_order.
  '''
  pass



class AsyncJob(object):
  '''
  Class to encapsulate asynchronous jobs.
  '''
  def __init__(
    self,
    func,
    args=None,
    kwargs=None,
    callback=None,
    error_callback=None,
    result_handler=None,
    error_handler=None
  ):
    '''
    func:
      The function to run.

    args:
      An iterable of the arguments to the function.

    kwargs:
      A dictionary fo the keyword arguments to the function.

    callback:
      The callback to run on the results of the function.

    error_callback:
      The error callback to run on an exception raised by the function.

    result_handler:
      A function to transform the result of the job. Called in parent thread
      with the result of the job and returns its return value.

    error_handler:
      A function to handle the error raised by the job. Called in parent thread
      with the exception raised by the job and returns its return value.
    '''
    self.func = func
    self.args = args
    self.kwargs = kwargs
    self.callback = callback
    self.error_callback = error_callback
    self.result = None
    self.result_handler = result_handler
    self.error_handler = error_handler


  def get_apply_async_args(self, n=None):
    '''
    Return arguments for multiprocessing.apply_async.
    '''
    if n is None:
      for n, item in zip(
        range(5,0,-1),
        (
          self.error_callback,
          self.callback,
          self.kwargs,
          self.args,
          self.func
        )
      ):
        if item is not None:
          break;
    yield self.func
    if n > 1:
      yield self.args if self.args is not None else tuple()
    if n > 2:
      yield self.kwargs if self.kwargs is not None else dict()
    if n > 3:
      yield self.callback
    if n > 4:
      yield self.error_callback


  def apply_async(self, pool):
    '''
    Pass the job to apply_async.
    '''
    self.result = pool.apply_async(*self.get_apply_async_args())


  def apply_sync(self):
    '''
    Run the job directly, i.e. not via the multiprocessing pool.
    '''
    func, args, kwargs = self.get_apply_async_args(n=3)
    try:
      result = func(*args, **kwargs)
    except Exception as e:
      return self.process_exception(e)
    else:
      return self.process_result(result)


  def process_result(self, result):
    '''
    Apply callback to result if there is one and return it.
    '''
    # Using a try-cache block here will catch TypeErrors in the callback.
    if self.result_handler is None:
      return result
    else:
      return self.result_handler(result)


  def process_exception(self, exception):
    '''
    A[[;u error callback to exception if there is one.
    '''
    # Using a try-cache block here will catch TypeErrors in the callback.
    if self.error_handler is None:
      raise exception
    else:
      return self.error_handler(exception)


  def get_async_result(self):
    '''
    Get the result from apply_async and process it before returning it.
    '''
    try:
      result = self.result.get()
    except Exception as e:
      return self.process_exception(e)
    else:
      return self.process_result(result)



def apply_async(
  jobs,
  max_workers=None
):
  '''
  Run the given jobs in parallel but invoke callbacks in the calling thread.

  jobs:
    A dictionary mapping identifiers to lists of AsyncJobs:

      {
        <id>: [<AsyncJob>, <AsyncJob>, <AsyncJob>],
        <id>: [<AsyncJob>, <AsyncJob>, <AsyncJob>]
      }


    To process a list of jobs without an ID, use apply_async_without_id.

  max_workers:
    Set the maximum number of works. If not set, this will be equal to
    os.cpu_count, or 2 if less than 1. If the number of jobs is less than
    max_workers, the number of jobs will be used.

  For each job, yield the identifier and the return value of the function.
  '''
  if not jobs:
    return

  n_jobs = sum(len(js) for js in jobs.values())

  if n_jobs < 1:
    return

  if max_workers is None:
    max_workers = max(os.cpu_count(), 2)
  else:
    max_workers = max(max_workers, 1)

  n_workers = min(n_jobs, max_workers)

  with mp.Pool(n_workers) as pool:
    queue = collections.deque()
    for ID, js in jobs.items():
      for job in js:
        job.apply_async(pool)
        queue.append((ID, job))
    while queue:
      ID, job = queue.popleft()
      # Allow jobs to launch other jobs. This can be used to try different
      # downloads in order, for example.
      result = job.get_async_result()
      if isinstance(result, AsyncJob):
        job.apply_async(pool)
        queue.append((ID, job))
      else:
        yield ID, result



def apply_async_without_id(jobs, *args, **kwargs):
  '''
  A wrapper around apply_async that accepts a list of jobs without an associated
  ID.
  '''
  for _, results in apply_async({None: jobs}, *args, **kwargs):
    yield results



def apply_async_with_grouped_jobs(
  grouped_jobs,
  *args,
  **kwargs
):
  '''
  A wrapper around apply_async where "grouped_jobs" is a map of identifiers to
  lists of dictionaries where each dictionary contains a list of jobs and a
  callback:

    {
      <id>: [
        {
          "jobs": [<AsyncJob>, <AsyncJob>, <AsyncJob>],
          "callback": <function>,
          "error_callback": <function>
        },
        {
          "jobs": [<AsyncJob>, <AsyncJob>, <AsyncJob>],
          "callback": <function>,
          "error_callback": <function>
        }
      ],
      <id>: [
        {
          "jobs": [<AsyncJob>, <AsyncJob>, <AsyncJob>],
          "callback": <function>,
          "error_callback": <function>
        },
        {
          "jobs": [<AsyncJob>, <AsyncJob>, <AsyncJob>],
          "callback": <function>,
          "error_callback": <function>
        }
      ]
    }
  For each dictionary, the job results will be collected in order and passed to
  the callback as a list. The return value of the callback will be yielded along
  with the ID. If the callback is None then the list of results will be yielded
  with the ID instead. The provided callback will be called in the invoking
  thread.

  Additional arguments and keyword arguments are passed through to apply_async.

  For example,

    {
      "foo": [
        {
          "jobs": [<AsyncJob foo1>, <AsyncJob foo2],
          "callback": <function f1>
        },
        {
          "jobs": [<AsyncJob fooA>, <AsyncJob fooB],
          "callback": None
        }
      ],
      "bar": [
        {
          "jobs": [<AsyncJob bar1>, <AsyncJob bar2],
          "callback": <function b1>
        }
      ]
    }

  will yield

    "foo", <function f1>((<results foo1>, <results foo2>))
    "foo", [<results fooA>, <results fooB>)]
    "bar", <function b1>((results bar1>, <results bar2>))
  '''
  groups = dict()
  joblist = dict()
  for ID, grps in grouped_jobs.items():
    for i, grp in enumerate(grps):
      grpID = (ID, i)
      jobs = grp['jobs']
      joblist.setdefault(grpID, []).extend(jobs)
      groups[grpID] = {
        'ID': ID,
        'callback': grp.get('callback', None),
        'n': len(jobs),
        'results': list()
      }
  for grpID, result in apply_async(joblist, *args, **kwargs):
    ID = groups[grpID]['ID']
    callback = groups[grpID]['callback']
    results = groups[grpID]['results']
    results.append(result)
    if len(results) ==  groups[grpID]['n']:
      del groups[grpID]
      if callback:
        yield ID, callback(results)
      else:
        yield ID, results



def attempt_jobs_in_order(jobs):
  '''
  A helper function for asynchronous jobs. Rather than try to run the list of
  jobs in parallel, they will be run sequentially until one succeeds.
  attempt_jobs_in_order can be passed as a job along with other asynchronous
  jobs to be run in parallel.

  jobs:
    A list of jobs where each job is the same described in apply_async.

      [<AsyncJob>, <AsyncJob>, <AsyncJob>]

    Error callbacks will be ignored. A job that fails should raise a JobError
    exception, either directly or via the callback.

  Returns the results of the first successful job, passed through the callback
  if given. Raises a JobError if no job completes successfully. Raises any other
  error raised by a job in the list. Jobs passed to this list should catch
  expected errors and wrap them in JobErrors.
  '''
  error = JobError('no jobs')
  for job in jobs:
    try:
      return job.apply_sync()
    except JobError as e:
      error = e
      continue
  raise error



################################ Miscellaneous #################################

def exhaust_generators(gen):
  '''
  Loop through all items in a generator to ensure complete processing and
  discard the output. If any of the items are themselves generators then
  recursively pass them to this function.
  '''
  for g in gen:
    if isinstance(g, types.GeneratorType):
      exhaust_generators(g)


def qualname(obj):
  '''
  Get the fully qualified class name of an object, e.g. Foo.Bar.Whatever
  '''
  return '{}.{}'.format(
    obj.__module__,
    obj.__class__.__qualname__
  )

