#!/usr/bin/env python3
# -*- coding: utf8 -*-

# Copyright (C) 2012-2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from collections import OrderedDict
import argparse
import collections
import colorsysplus
import copy
import functools
import glob
import importlib
import itertools
import json
import logging
import multiprocessing.dummy as mp
import os
import pwd
import pyalpm
import re
import shlex
import shutil
import stat
import subprocess
import sys

import Aria2JsonRpc
import XCGF
import XCPF.Pacman
import XCPF.PacmanRepoWrapper
import XCPF.RepoWrapper



################################## Constants ###################################

DOWNLOAD_LOCK_FILE = 'dl' + XCPF.Pacman.EXT_LOCK
SYSTEM_CONFIG_DIR = os.path.abspath(os.getenv('PB_SYSTEM_CONFIG_DIR', '/etc/pacboy'))



################################## Functions ###################################


def nested_update(a, b, deepcopy=False):
  '''
  Recursively update dictionary a with nested values from dictionary b.
  '''
  for k, v in b.items():
    if k in a and isinstance(a[k], dict) and isinstance(v, dict):
      nested_update(a[k], v, deepcopy=deepcopy)
    else:
      if deepcopy:
        try:
          v = copy.deepcopy(v)
        except AttributeError:
          pass
      a[k] = v
  return a



def recursively_load_config(path, base=None, loaded=None):
  '''
  Recursively load configuration files. The configuration file is a map in
  JSON format. The top-level key "base" may specify a path to load first before
  overriding it with the values in the current file.

  path:
    The path to load.

  base:
    An optional parsed base configuration object.

  loaded:
    A set of paths that have already been loaded, to prevent recursion.
  '''
  if loaded is None:
    loaded = set()
  if os.path.isabs(path):
    fallback_path = None
  else:
    fallback_path = os.path.join(SYSTEM_CONFIG_DIR, path)
  path = os.path.abspath(path)
  if path in loaded:
    return base
  loaded.add(path)
  try:
    try:
      with open(path, 'r') as f:
        obj = json.load(f)
    except (FileNotFoundError, json.decoder.JSONDecodeError) as e:
      raise PacboyError('failed to load {}'.format(path), error=e)
  except FileNotFoundError as e:
    if fallback_path is None:
      raise e
    else:
      return recursively_load_config(fallback_path, base=base, loaded=loaded)
  try:
    basepath = obj['base']
  except KeyError:
    pass
  else:
    if basepath:
      if not os.path.isabs(basepath):
        absbasepath = os.path.join(os.path.dirname(path), basepath)
        if not os.path.exists(absbasepath):
          absbasepath = os.path.join(SYSTEM_CONFIG_DIR, basepath)
      base = recursively_load_config(absbasepath, base=base, loaded=loaded)
  if base is not None:
    base = copy.deepcopy(base)
    obj = nested_update(base, obj)
  return obj



def diff_config(a, b):
  '''
  Delete redundant values from a that are supplied by b. Use this to minimize
  a configuration file (a) given a base (b).
  '''
  # list() is necessary to enable modification of a below.
  for k, v_a in list(a.items()):
    try:
      v_b = b[k]
    except KeyError:
      continue

    if isinstance(v_a, dict) and isinstance(v_b, dict):
      diff_config(v_a, v_b)
    elif v_a == v_b:
      del a[k]
  return a



def insert_config_base(conf, path, base=None):
  '''
  Insert the given path as a base for this configuration file and delete any
  values which are already supplied by the base.

  Returns the simplified configuration object. The original is not modified in
  situ.

  conf:
    The configuration object to simplify and in which to insert the path as
    base.

  path:
    The path to the base.

  base:
    The base of the base, if any. This defaults to Pacboy.DEFAULT_CONF.
  '''
  a = copy.deepcopy(conf)
  b = recursively_load_config(path, base=base)
  return diff_config(a, b)



def return_object_and_method_results(obj, method):
  '''
  Decorator to invoke an object method and return it along with the results of
  the method call.
  '''
  def func(*args, **kwargs):
    result = getattr(obj, method)(*args, **kwargs)
    return obj, result
  return func


############################ TemporarySyncDatabase #############################

class TemporarySyncDatabase():
  '''
  Context manager to temporary change the sync database while maintaining the
  same local database. This can be used to retrieve refreshed metadata without
  refresching the system.

  This will create a new database and symlink the local database into it. It
  will also update the database path in the nested pacman configuration file.
  These changes will be undone when the context exits.

  Packages must not be installed within the context as this would be a partial
  upgrade.
  '''
  def __init__(self, pacboy, dbpath=None):
    '''
    pacboy:
      Pacboy instance. This must already have its arguments and configuration
      file loaded and should be the instance returned by Pacboy's context
      switcher.

    dbpath:
      The temporary database path.
    '''
    self.pacboy = pacboy
    self.dbpath = dbpath
    self.orig_dbpath = None
    self.orig_value = None
    self.usage = dict()


  def __enter__(self):
    logging.debug('TemporarySyncDatabase: entering context with {}'.format(self.dbpath))
    logging.info('creating temporary database: {}'.format(self.dbpath))
    self.pacboy.disable_installation = True
    try:
      self.orig_value = self.pacboy.cpo.pacman_conf.options['DBPath']
    except KeyError:
      pass
    self.orig_dbpath = self.pacboy.cpo.pacman_conf.get_option('DBPath')

    dbpath = os.path.abspath(self.orig_dbpath)
    sync_path = os.path.join(dbpath, 'sync')
    local_path = os.path.join(dbpath, 'local')

    tmp_sync_path = os.path.join(self.dbpath, 'sync')
    tmp_local_path = os.path.join(self.dbpath, 'local')

    os.makedirs(tmp_sync_path, exist_ok=True)

    # Symlink the local database.
    XCGF.relink(local_path, tmp_local_path)

    # Copy the sync databases unless older than in the temporary path.
    glob_pattern = os.path.join(sync_path, '*')
    for db in glob.iglob(glob_pattern):
      name = os.path.basename(db)
      target = os.path.join(tmp_sync_path, name)
      XCGF.copy_if_absent_or_newer(db, target)

    self.pacboy.cpo.pacman_conf.options['DBPath'] = self.dbpath
    # Refresh the databases. This also reloads sync databases in the repo
    # wrappers.
    if self.pacboy.pargs.pb_temporary_database_refresh:
      self.pacboy.refresh(
        force=(self.pacboy.pargs.pb_temporary_database_refresh>1)
      )
    self.pacboy.cpo.refresh()

    return self.pacboy


  def __exit__(self, exc_type, exc_value, traceback):
    if self.orig_value is None:
      del self.pacboy.cpo.pacman_conf.options['DBPath']
    else:
      self.pacboy.cpo.pacman_conf.options['DBPath'] = self.orig_value
    self.pacboy.disable_installation = False
    logging.debug('TemporarySyncDatabase: leaving context with {}'.format(self.dbpath))




############################### Pacboy Arguments ###############################

def add_pacboy_arguments(parser):
  '''
  Add pacboy arguments to the argument parser.
  '''
  parser.add_argument(
    '--pb-build', metavar='<regex>', action='append', default=[],
    help='Additional build regular expressions.'
  )
  parser.add_argument(
    '--pb-nobuild', metavar='<regex>', action='append', default=[],
    help='Additional nobuild regular expressions.'
  )
  parser.add_argument(
    '--pb-build-all', '--pb-ba', action='store_true',
    help='Build all buildable packages.'
  )
  parser.add_argument(
    '--pb-build-directory', '--pb-bd', metavar='<path>', default='build',
    help='Build directory in which to download PKGBUILDs and source files and generate build scripts. Default: %(default)s.'
  )
  parser.add_argument(
    '--pb-clean', action='store_true',
    help='Clean up leftover Aria2 control files and exit.'
  )
  parser.add_argument(
    '--pb-config', metavar='<path>',
    help='The Pacboy configuration file that specifies repos, repo wrappers and repo extenders.',
  )
  parser.add_argument(
    '--pb-debug', action='store_true',
    help='Display debug messages. Use this instead of Pacman\'s --debug option to limit debug messages to %(prog)s.'
  )
  parser.add_argument(
    '--pb-disable', metavar='repo1,repo2,...',
    help='A comma-separated list of repos to disable for this operation. "*" will match all official databases. This will change the "Usage" settings of these repos. For finer control, change the configuration file. See the "Usage" section of the pacman.conf man page for help.'
  )
  parser.add_argument(
    '--pb-download-pkgbuilds', '--pb-dp', metavar='<path>',
    help='Download PKGBUILDs and related files to the given directory then exit.'
  )
  parser.add_argument(
    '--pb-download-sources', '--pb-ds', action='store_true',
    help='Attempt to download source files along with PKGBUILDs if possible. This is only possible if the repo(wrapper) supports it. This does NOT run makepkg or mksrcinfo on the downloaded PKGBUILDs.'
  )
  parser.add_argument(
    '--pb-no-repo-subdirs', '--pb-nrs', action='store_true',
    help='When downloading PKGBUILDs and related files, do not create pre-repo subdirectories in the download directory. Due to asynchronous parallel downloads, overlapping names will result in unexpected behavior. Use at your own risk.'
  )
  parser.add_argument(
    '--pb-export-default-config', '--pb-edc', metavar='<path>',
    help='Export the default configuration file to a path. If <path> is "-", print it to STDOUT.'
  )
  parser.add_argument(
    '--pb-export-config', '--pb-ec', metavar='<path>',
    help='Export the loaded configuration file to a path. If <path> is "-", print it to STDOUT.'
  )
  parser.add_argument(
    '--pb-gen-script', '--pb-gs', action='store_true',
    help='Generate the script and exit. See --pb-no-packages.'
  )
  parser.add_argument(
    '--pb-no-packages', '--pb-np', action='store_true',
    help='Do not download packages. Use to skip package downloads when generating build scripts. Implies --pb-gen-script.'
  )
  parser.add_argument(
    '--pb-no-checkdeps', '--pb-nc', action='store_true',
    help='Do not download checkdepends.'
  )
  parser.add_argument(
    '--pb-pre-install', '--pb-pi', action='store_true',
    help='Attempt to install all binary packages (including build targets, if available) before building packages. This may lead to redundant installations but it will avoid situations with intermediate partial upgrades during the build process.'
  )
  parser.add_argument(
    '--pb-rebuild-vcs', '--pb-rv', action='store_true',
    help='Build all VCS packages (Git, SVG, etc.). Which packages are recognized as VCS packages depends on the patterns in the configuration file.'
  )
  parser.add_argument(
    '--pb-snippet-help', '--pb-sh', action='store_true',
    help='Display build script snippet help.'
  )
  parser.add_argument(
    '--pb-threads', metavar='n', type=int, default=max(os.cpu_count(), 2),
    help='Set the maximum number of threads to use. Default: %(default)s'
  )
  parser.add_argument(
    '--pb-temporary-database', '--pb-td', metavar='<path>',
    help='Set a temporary database path to use to refresh sync databases instead of refreshing the system databases. This directory will also be used for other temporary metadata require for the operation that should not persist after it has completed.'
  )
  parser.add_argument(
    '--pb-temporary-database-refresh', '--pb-tdr', action='count',
    help='Refresh the temporary database when using --pb-tmpdb. This behaves the same way as pacman\'s -y/--refresh option. Pass it twice to force a refresh.'
  )
  parser.add_argument(
    '--pb-ttl', metavar='<seconds>', default=XCPF.common.DEFAULT_TTL, type=int,
    help='The time-to-live of cached metadata, in seconds. Default: %(default)s.'
  )
  parser.add_argument(
    '--pb-verbose', action='store_true',
    help='Display info messages messages. Use this instead of Pacman\'s --verbose option to limit verbosity to %(prog)s.'
  )
  return parser



################################# PacboyError ##################################

class PacboyError(Exception):
  '''
  Exceptions raised by the Pacboy class.
  '''

  def __init__(self, msg, error=None):
    '''
    Common error raised by Pacboy.

    msg: The error message.
    error: An Exception associated with the error, e.g. an HTTP error.
    '''
    self.msg = msg
    self.error = error

  def __str__(self):
    if self.error is None:
      return 'PacboyError: {}'.format(self.msg)
    else:
      return 'PacboyError: {} [{}]'.format(self.msg, self.error)



############################## Pacboy Decorators ###############################

def requires_downloader(func):
  '''
  Decorator to launch the Aria2 RPC server as needed for certain functions.
  '''
  def launching_func(self, *args, **kwargs):
    self.launch_aria2_rpc_server()
    func(self, *args, **kwargs)
  return launching_func



#################################### Pacboy ####################################

class Pacboy(object):
  '''
  Class to determine targets with dependency resolution for various package
  operations.
  '''
  DEFAULT_CONF = {
    'aria2': {
      'command': [
        '/usr/bin/aria2c',
        '--allow-overwrite=true',
        '--always-resume=false',
        '--auto-file-renaming=false',
        '--check-integrity=true',
        '--conditional-get=true',
        '--continue=true',
        '--file-allocation=none',
        '--log-level=error',
        '--max-concurrent-downloads=50',
        '--max-connection-per-server=5',
        '--min-split-size=5M',
        '--remote-time=true',
        '--rpc-listen-all=false',
        '--show-console-readout=false'
      ],
      'server': {'port': 6800, 'token': None},
    },
    'build': {
      'pre-install': False,
      'snippets': {
        'prologue': os.path.join(SYSTEM_CONFIG_DIR, 'snippets/prologue.sh'),
        'epilogue': None,
        'build': os.path.join(SYSTEM_CONFIG_DIR, 'snippets/build.sh'),
        "pacman": None,
        'custom build': []
      },
      'user': None,
      'VCS download command': [
        'makepkg-download_vcs_sources',
        '-o', '%OUTDIR%', '%URI%'
      ],
      'VCS patterns': [
        '-git$',
        '-hg$',
        '-svn$'
      ]
    },
    'env': {
      XCPF.RepoWrapper.ENV_KEY_TTL: XCPF.common.DEFAULT_TTL,
      XCPF.RepoWrapper.ENV_KEY_PACSERVE: None, #'http://localhost:15678',
      XCPF.RepoWrapper.ENV_KEY_IPFS: {
        'localhost': None, #'http://localhost:8080',
        'repos': {
          'core': None, #'/ipfs/QmU1qYFBYxvDSZPBAEL8ijKz4uA6kBVeRvexMWzHkytQvf/$repo/os/$arch'
        },
        'aria2 parameters': {
          #'timeout': '20'
        }
      }
    },
    'packages': {
      'build': [],
      'nobuild': [],
      'repo associations': []
    },
    'pacman': {
      'config': '/etc/pacman.conf',
      'command': ['/usr/bin/pacman']
    },
    'repos': {
      'enabled': ['*'],
      'repo wrapper factories': [],
      'build extender factories': [],
      'usage': {}
    },
    'sudo': {
      'command': ['/usr/bin/sudo']
    }
  }

  # Build prefixes.
  BUILD_PREFIX = '[+]'
  NOBUILD_PREFIX = '[-]'

  # Snippet interpolation variables and their descriptions.
  SNIPPET_VARIABLES = {
    'REPO': 'Package repository.',
    'PKGNAME': 'Package name.',
    'PKGBASE': 'Package base.',
    'VERSION': 'Package version.',
    'MAINTAINERS': 'Package maintainers or creators.',
    'RID': 'Release identifier.',
    'BUILDDIR': 'The build directory with the PKGBUILD.',
    'PACMAN_CONF': 'The pacman configuration file.',
    'IS_DEP': 'If the package is a dependency, insert the user-specified value of IS_DEP_TRUE, otherwise IS_DEP_FALSE.',
    'GRAPH_LEVEL': 'The level of this package in the dependency graph. All packages with the same level should be independent of each other. This can be used to build packages in parallel with clever snippets.',
    'COMMANDS': 'This is only used in "pacman" snippets. COMMANDS will be replaced by shell-escaped commands.',
  }

  def __init__(self, cpo=None, conf_base=None, conf_override=None):
    '''
    cpo:
      CommonPacmanOptions instance
    '''
    if cpo is None:
      cpo = XCPF.Pacman.CommonPacmanOptions()
    self.cpo = cpo
    self.conf = dict()
    self.conf_base = conf_base
    self.conf_override = conf_override
    self.repowrappers = OrderedDict()
    self.pargs = None
    # Aria2RpcServer object
    self.a2rs = None
    self.download = None
    self.env = dict()

    # Disable package installation. At first this was intended to be a Usage
    # override for TemporarySyncDatabase which would disable "Install" usage,
    # but that would disable information retrieval.
    self.disable_installation = False


  def get_id(self):
    '''
    Return an identifier for lock files etc.
    '''
    return os.path.basename(sys.argv[0])


  def parse_args(self, parser, args=None, add_args=True):
    '''
    Load command-line arguments using the given parser, the Pacboy parser and
    the appropriate Pacman parsers.
    '''
    if add_args:
      parser = add_pacboy_arguments(parser)
    pargs, cpo_pargs = self.cpo.parse_args(
      args=args, parent_parser=parser
    )

    # Allow pb_download_pkgbuilds to imply -S.
    if pargs.pb_download_pkgbuilds is not None \
    and getattr(self.cpo.pargs, 'operation', 'other') == 'other':
      if args is None:
        args = sys.argv[1:]
      args.insert(0, '-S')
      return self.parse_args(parser, args=args, add_args=False)

    # Allow --pb-no-packages to imply --pb-gen-script
    if pargs.pb_no_packages:
      pargs.pb_gen_script = True

    self.pargs = pargs
    return pargs


  def configure_logging(self):
    '''
    Set the appropriate logging level given the current arguments.
    '''
    level = None
    if self.pargs.pb_debug:
      level = logging.DEBUG
    # For some operations this may be None.
    elif self.cpo.pargs:
      if self.cpo.pargs.debug:
        level = logging.DEBUG
      elif self.cpo.pargs.quiet:
        level = logging.ERROR
      elif self.cpo.pargs.verbose or self.pargs.pb_verbose:
        level = logging.INFO
    XCGF.configure_logging(
      level=level,
      color=self.cpo.use_color(),
      clear_existing_handlers=True
    )


  def load_config(self, path=None, base=None):
    '''
    Load the PacBoy configuration file. This loads the class's default
    configuration file. If the instance's "conf_base" attribute it set it will
    be used to update the configuration object. If a base is passed then it too
    will be used to update the object. The resulting configuration is then used
    as a base when loading the path, if given. Finally, the object is updated
    with settings from the "conf_override" attribute, if set.
    '''
    if not path:
      path = self.pargs.pb_config
    self.conf = copy.deepcopy(self.DEFAULT_CONF)
    if self.conf_base is not None:
      nested_update(self.conf, self.conf_base)
    if base is not None:
      nested_update(self.conf, base)
    if path:
      self.conf = recursively_load_config(path, base=self.conf)
    if self.conf_override:
      nested_update(self.conf, self.conf_override)


  def configure_pkgs(self, pkgs):
    '''
    Apply various settings to packages.
    '''
    for pkg in pkgs:
      if pkg.upgrade_of:
        pkg.is_ignored = self.is_ignored(pkg.upgrade_of, alpm=True)
        pkg.is_dep = (pkg.upgrade_of.reason == 1)
      pkg.with_checkdeps = not self.pargs.pb_no_checkdeps


  def strip_build_prefix(self, arg):
    '''
    Strip build prefixes from the argument.

    If there was a build prefix, returns True or False along with the stripped
    argument, otherwise None with the unchanged argument.
    '''
    if arg.startswith(self.BUILD_PREFIX):
      arg = arg[len(self.BUILD_PREFIX):]
      build = True
    elif arg.startswith(self.NOBUILD_PREFIX):
      arg = arg[len(self.NOBUILD_PREFIX):]
      build = False
    else:
      build = None
    return build, arg


  def unstrip_build_prefix(self, build, arg):
    '''
    The reverse of strip_build_prefix().
    '''
    if build is None:
      return arg
    else:
      return (self.BUILD_PREFIX if build else self.NOBUILD_PREFIX) + arg


  def apply_repo_associations(self, args, only_repo=False):
    '''
    Add repo prefixes to the given argument as deteremined by the repo
    associations in the configuration file. Any argument with a repo prefix will
    be returned unchanged.

    If only_repo is True, iterate over the associated repos only, which may be
    None.
    '''
    assocs = list(
      (re.compile(pat), repo)
      for pat, repo in self.conf["packages"]["repo associations"]
    )
    for arg in args:
      build, stripped_arg = self.strip_build_prefix(arg)
      repo, name = XCPF.common.repo_and_package(stripped_arg)
      if not repo:
        nam, rel, ver = XCPF.common.split_version_requirement(name)
        for pattern, repo in assocs:
          if pattern.search(nam):
            if not only_repo:
              arg = self.unstrip_build_prefix(build, '{}/{}'.format(repo, name))
            break
      if only_repo:
        yield repo
      else:
        yield self.unstrip_build_prefix(build, arg)


  def parse_build_prefixes(self):
    '''
    Parse the build and nobuild argument prefixes to determine which targets
    should be built or not. Appropriate arguments will be added to the parsed
    command-line options. The CommonPacmanOptions arguments will be stripped.
    '''
    if not self.cpo.pargs:
      return
    pure_args = list()
    for arg in self.cpo.pargs.args:
      build, arg = self.strip_build_prefix(arg)
      pure_args.append(arg)
      if build is not None:
        # If there is no repo prefix then add it to restrict the pattern to
        # pkgname matching.
        if '/' in arg:
          prefix = ''
        else:
          prefix = '.*/'
        arg = re.escape(arg)
        r = '^{}{}$'.format(prefix, arg)
        if build:
          self.pargs.pb_build.append(r)
        else:
          self.pargs.pb_nobuild.append(r)
    self.cpo.pargs.args = pure_args


  def get_is_build_target_function(self):
    '''
    Compile the regular expressions and perform other actions to generate a
    function that accepts a qualified package name and returns a boolean
    indicating if it is a build target.
    '''
    build_res = XCGF.recursive_get(self.conf, ('packages', 'build'), [])
    nobuild_res = XCGF.recursive_get(self.conf, ('packages', 'nobuild'), [])

    build_res.extend(self.pargs.pb_build)
    nobuild_res.extend(self.pargs.pb_nobuild)

    build_res = list(re.compile(r) for r in build_res)
    nobuild_res = list(re.compile(r) for r in nobuild_res)

    def is_build_target(pkgname):
      build = False
      for r in build_res:
        if r.search(pkgname):
          build = True
          break
      if build:
        for r in nobuild_res:
          if r.search(pkgname):
            build = False
            break
      return build

    return is_build_target


  def actionable(self):
    '''
    Return True if the loaded arguments specify an action specific to Pacboy
    that cannot be handled by Pacman.
    '''
    if getattr(self.cpo.pargs, 'operation', 'other') == 'other':
      return False
    elif self.cpo.pargs.operation == 'query':
      # TODO
      # Refine this to operations that can actually be run.
      return self.cpo.pargs.upgrades
    else:
      return True


  def configure_aria2_rpc_server(self, force=False):
    '''
    Configure the Aria2RpcServer object.
    '''
    if self.a2rs:
      if not force:
        return
      self.a2rs.kill()
    cmd = self.conf['aria2']['command']
    port = self.conf['aria2']['server']['port']
    token = self.conf['aria2']['server']['token']
    if not token:
      token = Aria2JsonRpc.random_token(64)
    ID = self.get_id()
    lock = mp.Lock()
    quiet = self.cpo.pargs.quiet or not (self.cpo.pargs.verbose or self.pargs.pb_verbose)
    self.a2rs = Aria2JsonRpc.Aria2RpcServer(
      cmd, token, port, ID, lock=lock, nice=False, quiet=quiet
    )
    a2jr = self.a2rs.get_a2jr()
    self.download = self.a2rs.a2jr.polymethod_download_bool


  def launch_aria2_rpc_server(self):
    '''
    Launch the Aria2 RPC server if necessary. configure_aria2_rpc_server() must
    be run before invoking this method.
    '''
    self.a2rs.launch()


  def stop_aria2_rpc_server(self):
    '''
    Stop the Aria2 server.
    '''
    if self.a2rs:
      try:
        self.a2rs.kill()
      except Aria2JsonRpc.Aria2JsonRpcError as e:
        logging.warn(str(e))


  def configure_environment(self):
    '''
    Configure the environment that will be passed around between the
    RepoWrappers and BuildExtenders etc. This must configure all of the
    expected key constants in XCPF.RepoWrapper.
    '''
    self.env.clear()
    nested_update(self.env, self.conf['env'], deepcopy=True)
    self.env[XCPF.RepoWrapper.ENV_KEY_REFRESH] = list()
    self.env[XCPF.RepoWrapper.ENV_KEY_CPO] = self.cpo
    self.env[XCPF.RepoWrapper.ENV_KEY_DOWNLOADER] = self.download
    self.env[XCPF.RepoWrapper.ENV_KEY_IS_BUILD_TARGET] = self.get_is_build_target_function()
    self.env[XCPF.RepoWrapper.ENV_KEY_PACBOY_CONF] = copy.deepcopy(self.conf)
    self.env[XCPF.RepoWrapper.ENV_KEY_TTL] = self.pargs.pb_ttl
    logging.debug('environment: {}'.format(
      json.dumps(self.env, indent='  ', sort_keys=True, cls=XCGF.InformationJSONEncoder)
    ))


  def __enter__(self):
    self.load_config()
    if self.pargs.pb_rebuild_vcs:
      self.cpo.pargs.args.extend(self.rebuildable_installed_vcs_packages())
    if self.pargs.pb_build_all:
      self.pargs.pb_build = ['.*']
    if self.pargs.pb_download_pkgbuilds:
      self.pargs.pb_build = ['.*']
      self.pargs.pb_nobuild = []
    self.parse_build_prefixes()
    self.configure_usage()
    self.configure_aria2_rpc_server()
    self.configure_environment()
    self.load_repowrappers_from_conf()
    return self


  def __exit__(self, exc_type, exc_value, traceback):
    self.stop_aria2_rpc_server()


  def configure_usage(self):
    '''
    Configure repo usage from command-line parameters.
    '''
    if not self.pargs.pb_disable:
      return
    for disabled in self.pargs.pb_disable.split(','):
      for repo in self.cpo.expand_repos(disabled):
        self.conf['usage'][repo] = None


  def usage(self, repo, operation):
    '''
    Check the Usage level for this repo and return a boolean indicating if the
    requested operation is permitted.
    '''
    try:
      u = self.conf['usage'][repo]
    except (TypeError, KeyError):
      return self.cpo.usage(repo, operation)
    else:
      return (u and (operation in u or 'All' in u))


  def clear(self):
    '''
    Clear repo wrappers and related data.
    '''
    self.repowrappers.clear()


  def enabled_repos(self):
    '''
    If the configuration file specifies repos, iterate over them, otherwise
    iterate over the sync repos.
    '''
    return self.cpo.expand_repos(XCGF.recursive_get(
      self.conf, ('repos', 'enabled'), '*'
    ))


  def is_ignored(self, pkg, alpm=False):
    '''
    Determine if a Pkg or pyalpm package is ignored.
    '''
    if alpm:
      n = pkg.name
      gs = pkg.groups
    else:
      n = pkg.pkgname()
      gs = pkg.groups()
    return (
      self.cpo.is_ignored_pkg(n) or
      any(self.cpo.is_ignored_group(g) for g in gs)
    )


  def load_repowrappers_from_pacman(self):
    '''
    Load sync repos without build support, i.e. emulate Pacman.
    '''
    self.clear()
    for db in self.cpo.handle.get_syncdbs():
      self.repowrappers[db.name] = XCPF.PacmanRepoWrapper.PacmanRepoWrapper(
        self.env, db.name
      )


  def load_repowrappers_from_conf(self):
    '''
    Load repo wrappers and build extenders specified by the Pacboy configuration
    file.
    '''
    if not self.conf:
      return self.load_repowrappers_from_pacman()


    # Map of repos to RepoWrapperFactory subclasses.
    rw_factory_classes = dict()

    # Map of RepoWrapperFactory subclasses to instances.
    rw_factories = dict()

    # Map repos to classes.
    for x in XCGF.recursive_get(
      self.conf, ('repos', 'repo wrapper factories'), default=[]
    ):
      for repo in self.cpo.expand_repos(x[1:]):
        if repo not in rw_factory_classes:
          rw_factory_classes[repo] = x[0]

    # Default to PacmanRepoWrapperFactory
    for db in self.cpo.handle.get_syncdbs():
      if db.name not in rw_factory_classes:
        rw_factory_classes[db.name] = 'XCPF.PacmanRepoWrapper.PacmanRepoWrapperFactory'


    for repo in self.enabled_repos():
      try:
        rwf_cls = rw_factory_classes[repo]
      except KeyError:
        raise ValueError('{} repo is not enabled'.format(repo))
      else:
        try:
          rw_factory = rw_factories[rwf_cls]
        except KeyError:
          module, cls_name = rwf_cls.rsplit('.', 1)
          rw_factory = getattr(importlib.import_module(module), cls_name)()
          self.env = rw_factory.initialize(self.env)
          rw_factories[rwf_cls] = rw_factory
        self.env, self.repowrappers[repo] = rw_factory.wrap(self.env, repo)



  def download_syncdbs(self, force=False):
    '''
    Download sync databases files for all repos enabled in the currently loaded
    Pacman configuration file. Although this could be limited to repos specified
    in the Pacboy configuration file, the system would be left in a partially
    upgraded state if enabled repos vary between the two.
    '''
    # Callback to check that all required files were downloaded from a given
    # server.
    def check_download(dbname, server, required, results):
      if len(results) > len(required):
        zipper = itertools.zip_longest(required, results, fillvalue=False)
      else:
        zipper = zip(required, results)
      for req, success in zipper:
        if req and not success:
          raise XCGF.JobError(
            'failed to download database for {}'.format(dbname)
          )
      else:
        logging.info('downloaded {} from {}'.format(dbname, server))
        return dbname

    dbpath = os.path.abspath(self.cpo.pacman_conf.get_option('DBPath'))
    jobs = list()
    dbs = self.cpo.handle.get_syncdbs()
    for db in dbs:
      # These must be tried in order.
      names = list((db.name + x) for x in (
        XCPF.Pacman.EXT_DB, XCPF.Pacman.EXT_FILES_DB)
      )
      # Some repos don't provide the .files database.
      required = [True, False]
      # False, None or True
      siglevel = self.cpo.db_sigs(db.name)
      if siglevel is not False:
        # This was a fun bug. It creates an infinite loop that freezes the
        # entire system. Don't do this.
        # names.extend((n + XCPF.Pacman.EXT_SIG) for n in names)
        signames = list((n + XCPF.Pacman.EXT_SIG) for n in names)
        required.extend((bool(siglevel) and x) for x in required.copy())
        names.extend(signames)
      output_paths = list(os.path.join(dbpath, 'sync', n) for n in names)

      # The download jobs. These will be passed to XCGF.attempt_jobs_in_order to
      # try each server until a download succeeds.
      #
      # NB: The list of tuples ('addUri', ...) is the first argument, so it must
      # be wrapped in a tuple or list to represent the arguments passed to the
      # function.
      dl_jobs = list(
        XCGF.AsyncJob(
          self.download,
          # Arguments to download begin here.
          args=(
            # The first argument, which is the list of Aria2c downloads.
            list(
              (
                'addUri',
                ['{}/{}'.format(server.rstrip('/'), n)],
                {
                  'conditional-get': 'false' if force else 'true',
                  'out': os.path.basename(o),
                  'dir': os.path.dirname(o)
                }
              ) for (n, o) in zip(names, output_paths)
            ),
          ),
          # The callback.
          result_handler=functools.partial(check_download, db.name, server, required)
        ) for server in db.servers
      )

      # Append the jobs. Thanks to the callback, each one that succeeds will
      # return the database name.
      #
      # NB: dl_jobs is wrapped in a list it is the first argument to
      # attempt_jobs_in_order
      jobs.append(
        XCGF.AsyncJob(
          XCGF.attempt_jobs_in_order,
          args=(dl_jobs,)
        )
      )

    # Run the jobs.
    try:
      for dbname in XCGF.apply_async_without_id(jobs):
        pass
    except XCGF.JobError as e:
      raise RepoWrapper(str(e))


  def refresh(self, force=False, skip_sync_dbs=False):
    '''
    Refresh metadata. This will update the enabled pacman Sync databases, call
    all methods in the environment's ENV_KEY_REFRESH list and all refresh
    methods of the enabled RepoWrappers.

    The update order is the following:

    * Pacman sync databases.
    * CommonPacmanOptions objects.
    * ENV_KEY_REFRESH functions.
    * RepoWrappers.

    force:
      Force a refresh to emulate "pacman -Syy".
    '''
    if not skip_sync_dbs:
      dbpath, lockpath = self.cpo.db_path_and_lock()
      try:
        with XCGF.Lockfile(lockpath, 'database') as lock:
          self.cpo.refresh(force=force)
          self.download_syncdbs(force=force)
          self.cpo.refresh(force=force)
#           for rw in self.repowrappers.values():
#             if isinstance(rw, XCPF.PacmanRepoWrapper.PacmanRepoWrapper):
#               rw.load_db()
      except XCGF.LockError as e:
        logging.error(str(e))
        raise PacboyError('failed to lock {}'.format(dbpath), error=e)


    grouped_jobs = list()
    args = tuple()
    kwargs = {'force':force}

    # Refresh threadsafe shared objects.
    for func in self.env[XCPF.RepoWrapper.ENV_KEY_REFRESH]:
      grouped_jobs.extend(func(**kwargs))
    # It's a generator so cycle through it to make sure everything is run.
    for _ in XCGF.apply_async_with_grouped_jobs(
      {None : grouped_jobs}
    ):
      pass
    grouped_jobs.clear()

    # Refresh RepoWrappers.
    for rw in self.repowrappers.values():
      if not self.usage(rw.name, 'Sync'):
        continue
      grouped_jobs.extend(rw.refresh(**kwargs))
    # It's a generator so cycle through it to make sure everything is run.
    for _ in XCGF.apply_async_with_grouped_jobs(
      {None : grouped_jobs}
    ):
      pass


  def rebuildable_installed_vcs_packages(self):
    '''
    Return a set of names of installed VCS packages. This depends on the VCS
    patterns in the configuration file. The returned names will include the
    build prefix to ensure that they are built.
    '''
    patterns = list(
      re.compile(pat) for pat in self.conf['build']['VCS patterns']
    )
    targets = set()
    for pkg in self.cpo.handle.get_localdb().pkgcache:
      for pat in patterns:
        if pat.search(pkg.qualified_name):
          targets.add('{}{}'.format(BUILD_PREFIX, pkg.name))
          break
    return targets()


  def upgradable(self, targets=None, downgradable=False):
    '''
    Determine upgradable packages.

    If downgradable is True, return downgrade targets to emulate pacman -Suu.

    Returns a PkgSet() of upgradable packages and a set of strings of
    apparent orphans. The returned packages will also have the following
    attributes set:

    * upgrade_of: The pyalpm local package that the package is upgrading.
    * is_ignored: A boolean indicating if this upgrade should be ignored.
    * is_dep: A boolean indicating if the local package is a dependency.


    If targets are specified but do not correspond to installed packages then
    they will be ignored with a warning.
    '''
    local_db = self.cpo.handle.get_localdb()

    if targets:
      if not isinstance(targets, set):
        targets = set(targets)
      installed = list()
      for t in sorted(targets):
        l = local_db.get_pkg(t)
        if l:
          installed.append(l)
        else:
          logging.warning('{} is not an installed package'.format(t))
    else:
      installed = local_db.pkgcache

    if self.cpo.pargs.operation == 'query':
      # Apply filters
      if self.cpo.pargs.foreign:
        installed = (i for i in installed if not XCPF.common.find_pkgname(
          self.cpo.handle.get_syncdbs(), i.name
        ))
      elif self.cpo.pargs.native:
        installed = (i for i in installed if XCPF.common.find_pkgname(
          self.cpo.handle.get_syncdbs(), i.name
        ))

      if self.cpo.pargs.deps:
        installed = (i for i in installed if i.reason == 1)
      elif self.cpo.pargs.explicit:
        installed = (i for i in installed if i.reason == 0)

    installed = installed if isinstance(installed, list) else list(installed)


    installed_names = list(i.name for i in installed)
    # Ensure that packages are updated from the desired repo. This maps
    # package names to repo names or None.
    repo_associations = dict(zip(
      installed_names,
      self.apply_repo_associations(installed_names, only_repo=True)
    ))
    installed = dict((i.name, i) for i in installed)
    orphans = set(installed)
    upgradable = XCPF.Pkg.PkgSet()


    if installed:
      for repowrapper in self.repowrappers.values():
        if not self.usage(repowrapper.name, 'Upgrade'):
          continue
        check = list(
          installed[name] for name in orphans
          if repo_associations[name] in (None, repowrapper.name)
        )
        up, not_found = repowrapper.upgradable(check, downgradable=downgradable)
        upgradable |= up
        # Remove packages that were found in this repo. Orphans cannot be
        # updated directly from "up" or "upgradable" because upgrades may be
        # replacements with different names.
        orphans -= (set(p.name for p in check) - set(p.name for p in not_found))

      self.configure_pkgs(upgradable)

    return upgradable, orphans


  def display_upgradable(self, *args, **kwargs):
    '''
    Print upgradable packages to STDOUT. This attempts to emulate the way that Pacman
    displays upgradable packages.

    This accepts the same arguments as upgradable().
    '''
    if self.cpo.use_color():
      vc = colorsysplus.ansi_sgr(fg=10)
      reset = colorsysplus.ansi_sgr(reset=True)
    else:
      vc = ''
      reset = ''
    upgradable, orphans = self.upgradable(*args, **kwargs)
    for p in sorted(upgradable, key=lambda x: x.upgrade_of.name):
      upgrade_of = p.upgrade_of
      n1 = upgrade_of.name
      v1 = upgrade_of.version
      ignored = self.is_ignored(upgrade_of, alpm=True)
      if p.pkgname() == n1:
        n2 = ''
      else:
        n2 = p.pkgname() + ' '
      v2 = p.version()
      if ignored:
        i = ' [ignored]'
      else:
        i = ''
      print(
        '{n1} {vc}{v1}{reset} -> {n2}{vc}{v2}{reset}{i}'.format(
          n1=n1,
          n2=n2,
          v1=v1,
          v2=v2,
          vc=vc,
          reset=reset,
          i=i
        )
      )


  def satisfy(self, targets, *args, only_first=True, usage='Install', **kwargs):
    '''
    Satisfy the targets using the current repos, in order.

    targets:
      The possibly qualified package and group name targets to satisfy.

    only_first:
      If True, once a target has been satisfied by a repo, it will be removed
      from queries to subsequent repos. This will be True for e.g. package
      installation operations but False for queries.

    usage:
      The usage (as specified by the pacman.conf Usage parameter) for which
      the method has been called. For example, this my be set to 'Search' to
      limit the returned packages to those which should appear in a search.

    This also accepts additional arguments and keyword arguments recognized by
    XPCF.Repo.RepoWrapper.satisfy and returns the same values.
    '''
    satisfied = XCPF.Pkg.PkgSet()
    groups = dict()
    targets = set(self.apply_repo_associations(targets))
    unsatisfied = targets.copy()

    for repowrapper in self.repowrappers.values():
      if not self.usage(repowrapper.name, usage):
        continue
      if only_first and targets:
        if not unsatisfied:
          break
        # Each subsequent query only looks for unsatisfied targets.
        rsat, rgrp, unsatisfied = repowrapper.satisfy(
          unsatisfied, *args, **kwargs
        )
      else:
        # Each subsequent query looks for all targets.
        rsat, rgrp, runsat = repowrapper.satisfy(
          targets, *args, **kwargs
        )
        # The same targets are passed to the satisfy method each time. The
        # unsatisfied targets are those that were not satisfied by any
        # invocation, i.e. the intersection of the unsatisfied sets.
        unsatisfied &= runsat
      satisfied |= rsat
      for name, pkgs in rgrp.items():
        try:
          groups[name] |= pkgs
        except KeyError:
          groups[name] = pkgs

    self.configure_pkgs(itertools.chain(satisfied, *groups.values()))

    return satisfied, groups, unsatisfied


  def satisfy_dependencies(self, pkgs):
    '''
    Resolve dependencies for the given packages.

    Returns a PkgSet of satisfiers and a set of strings of unsatisfied
    dependencies. The satisfiers will have their "is_dep" attribute set to True.
    '''
    satisfiers = XCPF.Pkg.PkgSet()
    unsatisfied = set()
    installed = self.cpo.handle.get_localdb().pkgcache
    seen = set(d
      for pkg in pkgs
      for d in pkg.operation_deps()
    )
    deps = seen.copy()
    # Consider all targets to be virtually installed.
    virtually_installed = dict(
      (pkg.pkgname(), pkg.version()) for pkg in pkgs
    )
    # Handle Pacman's --assume-installed arguments.
    for vi in self.cpo.pargs.assume_installed:
      # This should always be of the form <pkgname>=<version>.
      n, _, v = split_version_requirement(vi)
      virtually_installed[n] = v

    while deps:
      unresolved = set()
      for d in deps:
        if pyalpm.find_satisfier(installed, d) \
        or XCPF.common.satisfied_by_names_and_versions(virtually_installed, d):
          continue
        unresolved.add(d)
      if not unresolved:
        break

      sat, _, unsat = self.satisfy(unresolved)
      satisfiers.update(sat)
      unsatisfied.update(unsat)
      virtually_installed = dict((pkg.pkgname(), pkg.version()) for pkg in sat)

      deps = set(d
        for pkg in sat
        for d in pkg.operation_deps()
        if d not in seen
      )
      seen.update(deps)

    for pkg in satisfiers:
      pkg.is_dep = True
    return satisfiers, unsatisfied


  def _error_template(self, msg):
    '''
    Emulate Pacman error messages.
    '''
    if self.cpo.use_color():
      prefix = '{}error:{} '.format(
        colorsysplus.ansi_sgr(fg=1),
        colorsysplus.ansi_sgr(reset=True)
      )
    else:
      prefix = 'error: '
    return prefix + msg


  def display_info(self, pkgnames, *args, extended=False, **kwargs):
    '''
    Print package information to STDOUT. This attempts to emulate the way that
    pacman displays package information.

    This accepts package names.
    '''
    if not (isinstance(pkgnames, list) or isinstance(pkgnames, tuple)):
      pkgnames = tuple(pkgnames)

    wrapper = XCPF.common.InfoDisplay(*args, **kwargs)

    satisfied, groups, unsatisfied = self.satisfy(
      pkgnames, only_first=False, usage='Search'
    )
    if not pkgnames:
      pkgnames = sorted(satisfied.names())
    for r in self.repowrappers:
      if not self.usage(r, 'Search'):
        continue
      for p in pkgnames:
        repo, name = XCPF.common.repo_and_package(p)
        if repo:
          if repo != r:
            continue
          else:
            n = p
        else:
          n = '{}/{}'.format(r, p)
        try:
          pkg = satisfied[n]
        except KeyError:
          for gs in groups.values():
            try:
              pkg = gs[n]
            except KeyError:
              continue
            else:
              break
          else:
            continue
        wrapper.display(pkg.display_fields(extended=extended))
        print()

    error_template = self._error_template('package \'{}\' was not found')
    for p in unsatisfied:
      print(error_template.format(p))


  def search(self, terms, by=XCPF.RepoWrapper.SEARCH_BYS[0], intersect=False):
    '''
    Search for packages.
    '''
    for rw in self.repowrappers.values():
      if not self.usage(rw.name, 'Search'):
        continue
      yield from rw.search(terms, by=by, intersect=intersect)


  def display_search(self, terms, *args, **kwargs):
    '''
    Print search results to STDOUT. This attempts to emulate the way that Pacman
    displays search results.

    This accepts the same arguments as search().
    '''
    if not terms:
      self.display_list(emulate_search=True)
      return

    if self.cpo.quiet():
      for p in self.search(terms, *args, **kwargs):
        print(p.pkgname())
    else:
      local_db = self.cpo.handle.get_localdb()
      if self.cpo.use_color():
        rc = colorsysplus.ansi_sgr(fg=13)
        vc = colorsysplus.ansi_sgr(fg=10)
        gc = colorsysplus.ansi_sgr(fg=12)
        ic = colorsysplus.ansi_sgr(fg=14)
        reset = colorsysplus.ansi_sgr(reset=True)
      else:
        rc = ''
        vc = ''
        gc = ''
        ic = ''
        reset = ''
      for p in self.search(terms, *args, **kwargs):
        grps = sorted(p.groups())
        if grps:
          grpinfo = '{} ({})'.format(gc, ' '.join(grps))
        else:
          grpinfo = ''
        if local_db.get_pkg(p.pkgname()):
          installed_info = ic + ' [installed]'
        else:
          installed_info = ''
        print(
          '{rc}{repo}/{reset}{name} {vc}{ver}{grp}{inst}{reset}\n  {desc}'.format(
            repo=p.repo(),
            name=p.pkgname(),
            ver=p.version(),
            grp=grpinfo,
            inst=installed_info,
            desc=p.pkgdesc(),
            rc=rc,
            vc=vc,
            reset=reset
          )
        )


  def list(self, repos=None):
    '''
    List packages in the given repos, or all if none are given. This will
    iterate over tuples of

        <repo>, <name>, <version>, <description>

    where <version> and <description> may be None and where each repo may
    yield an empty list if returning the full list would involve unreasonable
    overhead.

    repos:
      A list of repos to list.
    '''
    if not repos:
      repos = self.repowrappers.keys()
    else:
      error_template = self._error_template('repository "{}" was not found.')
      rs = list()
      for r in repos:
        if r not in self.repowrappers:
          print(error_template.format(r))
        else:
          rs.append(r)
      repos = rs

    for r in repos:
      if not self.usage(r, 'Search'):
        continue
      for n, v, d in self.repowrappers[r].list():
        yield r, n, v, d


  def display_list(self, *args, emulate_search=False, **kwargs):
    '''
    Print lists to STDOUT. This attempts to emulate the way that Pacman displays
    lists.

    This accepts the same arguments as list().
    '''
    if self.cpo.quiet():
      for _, n, _, _ in self.list(*args, **kwargs):
        print(n)
      return


    local_db = self.cpo.handle.get_localdb()
    if self.cpo.use_color():
      rc = colorsysplus.ansi_sgr(fg=13)
      vc = colorsysplus.ansi_sgr(fg=10)
      ic = colorsysplus.ansi_sgr(fg=14)
      reset = colorsysplus.ansi_sgr(reset=True)
    else:
      rc = ''
      vc = ''
      ic = ''
      reset = ''
    for r, n, v, d in self.list(*args, **kwargs):
      if not v:
        v = ''
      else:
        v = vc + v
      if local_db.get_pkg(n):
        i = ic + ' [installed]'
      else:
        i = ''
      desc = ''
      if emulate_search:
        sep = '/'
        if desc:
          desc = '\n    {}'.format(d)
      else:
        sep = ' '
      print(
        '{rc}{r}{reset}{sep}{n} {v}{i}{reset}{desc}'.format(
          r=r,
          n=n,
          v=v,
          i=i,
          rc=rc,
          reset=reset,
          sep=sep,
          desc=desc
        )
      )


  def groups(self, names=None, include_members=False):
    '''
    Search for packages.
    '''
    if not (names or include_members):
      groups = set()
      is_set = True
    else:
      groups = dict()
      is_set = False
    for rw in self.repowrappers.values():
      if not self.usage(rw.name, 'Search'):
        continue
      results = rw.groups(names=names, include_members=include_members)
      if is_set:
        groups.update(results)
      else:
        for k, v in results.items():
          try:
            groups[k].update(v)
          except KeyError:
            groups[k] = v
    return groups


  def display_groups(self, *args, **kwargs):
    '''
    Display groups. Emulate pacman -Sg. All arguments are forwarded to groups().
    '''
    groups = self.groups(*args, **kwargs)
    if isinstance(groups, set):
      for g in sorted(groups):
        print(g)
    else:
      for g in sorted(groups):
        for p in sorted(groups[g].names()):
          print(g, p)


  def select_group_members(self, groups):
    '''
    Prompt the user to select group members.

    groups:
      A dictionary mapping group names to PkgSets.

    Iterate over the select packages.
    '''
    for g in sorted(groups):
      yield from XCPF.Pacman.select_grp(
        g, groups[g], ignored=self.cpo.ignored_pkgs()
      )


  def print(self, pkgs):
    '''
    Emulate Pacman's --print behavior. For build-only targets some data will
    not be available.
    '''
    fmt = self.cpo.pargs.print_format
    if not fmt:
      fmt = '%l'
    for pkg in sorted(pkgs, key=lambda x: x.pkgname()):
      if isinstance(pkg, XCPF.Pkg.PyalpmPkg):
        location = '{}/{}'.format(
          pkg.pkg.db.servers[0].rstrip('/'),
          pkg.pkg.filename
        )
        size = str(pkg.pkg.size)
      else:
        location = '?'
        size = '?'
      line = fmt
      for esc, val in (
        ('%n', pkg.pkgname()),
        ('%v', pkg.version()),
        ('%l', location),
        ('%r', pkg.repo()),
        ('%s', size)
      ):
        line = line.replace(esc, val)
      print(line)


  @staticmethod
  def partition_buildable(pkgs):
    '''
    Partition a PkgSet into a PkgSet of packages to build and a Pkgset of
    packages to install with Pacman.
    '''
    build = XCPF.Pkg.PkgSet(p for p in pkgs if p.build)
    nobuild = XCPF.Pkg.PkgSet(p for p in pkgs if not p.build)
    return build, nobuild


  def download_packages(self, outdir, pkgs):
    '''
    Invoke each RepoWrapper's download_packages method. This accepts the same
    arguments and returns the same values.
    '''
    jobs = list()
    found = XCPF.Pkg.PkgSet()
    by_repo = dict()

    for p in pkgs:
      by_repo.setdefault(p.repo(), list()).append(p)

    for rw in self.repowrappers.values():
      try:
        ps = by_repo[rw.name]
      except KeyError:
        continue
      else:
        r_found, r_jobs = rw.download_packages(outdir, ps)
        found |= r_found
        jobs.extend(r_jobs)

    return found, jobs


  def download_pkgbuilds(self, outdir, pkgs, with_source_files=False):
    '''
    Invoke each RepoWrapper's download_pkgbuilds method. This accepts the same
    arguments and returns the same values.
    '''
    jobs = list()
    by_repo = dict()
    by_pkgbase = dict()
    found = set()

    for p in pkgs:
      pkgbase = p.pkgbase()
      if pkgbase is NotImplemented:
        logging.warning('skipping unbuildable package {} (no pkgbase)'.format(
          p.qualified_pkgname())
        )
        continue
      by_pkgbase.setdefault(p.qualified_pkgbase(), XCPF.Pkg.PkgSet()).add(p)
      by_repo.setdefault(p.repo(), set()).add(pkgbase)

    for rw in self.repowrappers.values():
      rw_outdir = outdir if self.pargs.pb_no_repo_subdirs else os.path.join(outdir, rw.name)
      try:
        ps = by_repo[rw.name]
      except KeyError:
        continue
      else:
        r_found, r_jobs = rw.download_pkgbuilds(
          rw_outdir, ps,
          with_source_files=with_source_files
        )
        for p in r_found:
          found |= by_pkgbase[p]
        jobs.extend(r_jobs)

    return found, jobs


  def download_packages_and_pkgbuilds(self, pkgs, download_construenda=False):
    '''
    Download packages and PKGBUILDs for the given packages. What is downloaded
    depends on each package's "build" attribute.

    pkgs:
      PkgSet to download.

    download_construenda:
      Retrieve binary packages even for those that will be built. Construendus
      is the future passive participle of construo which means "that which is to
      be built". The nominative neutral pleural is "construenda" and means
      "the things which are to be built". It provides a very concise variable
      name.

    Returns a PkgSet for which the downloads were successful and a PkgSet for
    which the downloads.
    '''
    found = XCPF.Pkg.PkgSet()

    cachedir, cachelock = self.cpo.cache_path_and_lock()
    pkgbuild_dir = self.pargs.pb_build_directory
    pkgbuild_lock = os.path.join(pkgbuild_dir, DOWNLOAD_LOCK_FILE)

    with_source_files = self.pargs.pb_download_sources

    max_threads = self.pargs.pb_threads
    if max_threads is not None:
      max_threads = max(0, max_threads)

    build_pkgs, bin_pkgs = self.partition_buildable(pkgs)
    bin_pkg_map = dict((p.qualified_pkgname(), p) for p in bin_pkgs)
    build_pkg_map = dict()
    for p in build_pkgs:
      build_pkg_map.setdefault(p.qualified_pkgbase(), []).append(p)

    grouped_jobs = dict()
    PKG_JOBS = 0
    PKGBUILD_JOBS = 1
    PRE_INSTALL_PKG_JOBS = 2

    if bin_pkgs and not self.pargs.pb_no_packages:
      fnd, jbgrps = self.download_packages(cachedir, bin_pkgs)
      found |= fnd
      grouped_jobs[PKG_JOBS] = jbgrps


    if build_pkgs:
      # Get the PKGBUILD download jobs.
      fnd, jbs = self.download_pkgbuilds(
        pkgbuild_dir, build_pkgs, with_source_files=with_source_files
      )
      grouped_jobs[PKGBUILD_JOBS] = jbs
      for pkgbase in fnd:
        found.update(build_pkg_map[pkgbase])

      # Download
      if download_construenda and not self.pargs.pb_no_packages:
        _, jbs = self.download_packages(cachedir, build_pkgs)
        grouped_jobs[PRE_INSTALL_PKG_JOBS] = jbs

    ID = self.get_id()
    # Disable lock file if there is nothing to download in the directory.
    if grouped_jobs.get(PKG_JOBS, None) \
    or grouped_jobs.get(PRE_INSTALL_PKG_JOBS, None):
      os.makedirs(cachedir, exist_ok=True)
    else:
      cachelock = None
    if grouped_jobs.get(PKGBUILD_JOBS, None):
      os.makedirs(pkgbuild_dir, exist_ok=True)
    else:
      pkgbuild_lock = None

    try:
      with XCGF.Lockfile(cachelock, ID):
        with XCGF.Lockfile(pkgbuild_lock, ID):
          for category, results in XCGF.apply_async_with_grouped_jobs(
            grouped_jobs, max_workers=self.pargs.pb_threads
          ):
            # Ignore the PRE_INSTALL_PKG_JOBS as they are covered by PKGBUILD_JOBS.
            if category == PKG_JOBS:
              found.update(bin_pkg_map[r] for r in results)
            elif category == PKGBUILD_JOBS:
              for r in results:
                found.update(build_pkg_map[r])
    except XCGF.LockError as e:
      logging.error(str(e))
      raise PacboyError(
        'failed to lock {} or {}'.format(cachedir, pkgbuild_dir),
        error=e
      )

    self.chown_for_builder(pkgbuild_dir)
    return found


  def operation_download_pkgbuilds(self, pkgs):
    '''
    Download PKGBUILDs and related files according to the given options. This
    implements the operation specified by --pb-download-pkgbuilds.

    Returns a boolean indicating if the operation was successful (all targets
    downloaded).
    '''
    outdir = os.path.abspath(self.pargs.pb_download_pkgbuilds)
    os.makedirs(outdir, exist_ok=True)
    found, jobs = self.download_pkgbuilds(
      outdir,
      pkgs,
      with_source_files=self.pargs.pb_download_sources
    )
    for _, pkgbases in XCGF.apply_async_with_grouped_jobs(
      {None:jobs},
      max_workers=self.pargs.pb_threads
    ):
      pkgbases = list(pkgbases)
      found.update(pkgbases)

    basemap = dict()
    for p in pkgs:
      basemap.setdefault(p.qualified_pkgbase(), []).append(p.qualified_pkgname())
    for pkgbase in sorted(found):
      pkgnames = sorted(basemap[pkgbase])
      # Display pkgnames per pkgbase if multiple or different from pkgbase.
      if not pkgnames or (len(pkgnames) == 1 and pkgnames[0] == pkgbase):
        logging.info('retrieved pkgbase {}'.format(pkgbase))
      else:
        logging.info('retrieved pkgbase {} [{}]'.format(
          pkgbase, ' '.join(sorted(basemap[pkgbase]))
        ))

    not_found = pkgs.copy()
    not_found.remove_many(p for p in pkgs if p.qualified_pkgbase() in found)
    if not_found:
      logging.error('failed to retrieve PKGBUILDs for {}'.format(
        ' '.join(sorted(p.qualified_pkgname() for p in not_found))
      ))
    self.chown_for_builder(outdir)
    return (not not_found)


  @staticmethod
  def structure_installation(pkgs):
    '''
    Structure the installation procedure. Determine the dependency graph and
    order the packages accordingly. Iterate over lists of binary packages that
    can be installed with pacman and single packages that must be built.
    '''
    bin_pkgs = list()
    # Iterate over DependencyGraphNodes.
    for dgn in XCPF.Pkg.DependencyGraph(pkgs):
      if dgn.pkg.build:
        if bin_pkgs:
          yield bin_pkgs
          # Create a new list instead of clearing the one that was just yielded.
          bin_pkgs = list()
        yield dgn
      else:
        bin_pkgs.append(dgn)
    if bin_pkgs:
      yield bin_pkgs


  @classmethod
  def print_snippet_interpolation_help(cls):
    '''
    Print a help message with information about snippet interpolation. This
    could be moved to a man page but here the
    '''
    print('''Snippets take the following format:

    DELIMITER_LEFT=%%
    DELIMITER_RIGHT=%%
    IS_DEP_TRUE=--asdeps
    IS_DEP_FALSE=--asexplicit
    BEGIN
    #################################### Build #####################################
    # Target: %%REPO%%/%%PKGNAME%%-%%VERSION%%
    # Package Base: %%PKGBASE%%
    # Maintainer(s): %%MAINTAINERS%%
    # Release Identifier: %%RID%%
    pushd %%QUOTED_BUILDDIR%%
    pb-query_trust --rid %%QUOTED_RID%% %%QUOTED_REPO%%/%%QUOTED_PKGBASE%% %%QUOTED_MAINTAINERS%%
    makepkg -irs --pconfig %%QUOTED_PACMAN_CONF%% --pkg %%QUOTED_PKGNAME%% %%IS_DEP%%
    pacman --config %%QUOTED_PACMAN_CONF%% -T %%QUOTED_PKGNAME%% || exit 1
    popd

Everything up to and including the line with "BEGIN" is the header. This must
be included in every snippet. The lines preceding "BEGIN" configure values for
variable interpolation in the snippet and take the form <name>=<value>, with no
spaces. The following names are recognized:

  DELIMITER_LEFT & DELIMITER_RIGHT:
    This will set the left and right variable delimiters. In this example it is
    "%%" for both so '%%REPO%% will be recognized as a variable and replaced
    with the package repository when interpolating this snippet. See below for
    recognized variables.

  IS_DEP_TRUE:
    When the target package is a dependency, the variable "IS_DEP" will be
    replaced with this value.

  IS_DEP_FALSE:
    When the target package is not a dependency, the variable "IS_DEP" will be
    replaced with this value.

Everything after "BEGIN" is arbitrary and will be inserted in the output script
as-is after variable interpolation. If the snippet changes the working directory
then it should return to the initial directory when done.

The following variables are recognized:
{}

Each one may be prefixed with "QUOTED_" to ensure that the replacement is
shell-escaped.
'''.format(
      '\n'.join(
        '  {}:\n    {}'.format(k, v)
        for k, v in sorted(cls.SNIPPET_VARIABLES.items())
      )
    ))


  def interpolate_snippet(self, snippet, dgn=None, cmds=None):
    '''
    Interpolate variables in snippet with package values.

    snippet:
      A block of text with variables to interpolate.

    dgn:
      An XCPF.Pkg.DependencyGraphNode.

    cmd:
      Commands to interpolate in the snippet. This should be an iterable of
      strings and iterables. Strings are treated as per-formatted commands and
      inserted on a line verbatim. Iterables are treated as words and joined
      after shell-escaping.
    '''
    # Format the command block.
    commands = '\n'.join(
      cmd if isinstance(cmd, str) else XCGF.sh_quote_words(cmd) for cmd in cmds
    ) if cmds else ''
    if snippet is None:
      return '\n{}\n'.format(commands)

    pkg = None if dgn is None else dgn.pkg
    user_vars = dict()
    lines = iter(snippet.split('\n'))
    block = None
    for line in lines:
      if line == 'BEGIN':
        block = '\n'.join(lines)
        break
      else:
        k, v = line.split('=', 1)
        k = k.strip()
        v = v.strip()
        user_vars[k] = v
    else:
      raise PacboyError('not BEGIN line in snippet:\n{}'.format(snippet))

    if not block:
      raise PacboyError('empty snippet')
    try:
      delimiter_left = user_vars['DELIMITER_LEFT']
      delimiter_right = user_vars['DELIMITER_RIGHT']
      if pkg:
        is_dep = user_vars['IS_DEP_TRUE'] if pkg.is_dep else user_vars['IS_DEP_FALSE']
    except KeyError as e:
      raise PacboyError(
        'required user variable not set in snippet [{}]:\n{}'.format(e, snippet)
      )


    var_vals = list((
      ('PACMAN_CONF', self.conf['pacman']['config']),
      ('COMMANDS', commands),
    ))
    if pkg is not None:
      var_vals.extend((
        ('REPO', pkg.repo()),
        ('PKGNAME', pkg.pkgname()),
        ('PKGBASE', pkg.pkgbase()),
        ('VERSION', pkg.version()),
        ('MAINTAINERS', pkg.maintainers()),
        ('RID', pkg.rid()),
        ('BUILDDIR', os.path.join(pkg.repo(), pkg.pkgbase())),
        ('IS_DEP', is_dep),
        ('GRAPH_LEVEL', str(dgn.n)),
      ))

    for var, val in var_vals:
      if isinstance(val, list):
        quoted_val = ' '.join(shlex.quote(v) for v in val)
        val = ' '.join(val)
      else:
        quoted_val = shlex.quote(val)
      block = block.replace(
        '{}{}{}'.format(delimiter_left, var, delimiter_right),
        val
      )
      block = block.replace(
        '{}QUOTED_{}{}'.format(delimiter_left, var, delimiter_right),
        quoted_val
      )

    return block


  def determine_sync_targets(self):
    '''
    Determine the sync targets for the given arguments.
    '''
    pkgs = XCPF.Pkg.PkgSet()

    if self.cpo.pargs.sysupgrade:
      upgradable, orphans = self.upgradable(downgradable=self.cpo.pargs.sysupgrade>1)
      if orphans:
        logging.debug('orphans: {}'.format(' '.join(sorted(orphans))))
      pkgs |= upgradable

    if self.cpo.pargs.args:
      satisfied, groups, unsatisfied = self.satisfy(
        self.cpo.pargs.args,
        only_first=True,
        usage='Search'
      )
      if unsatisfied:
        logging.error('unrecognized targets: {}'.format(' '.join(sorted(unsatisfied))))
        sys.exit(1)
      pkgs |= satisfied
      if self.cpo.pargs.noconfirm:
        for ps in groups.values():
          pkgs |= ps
      else:
        pkgs.update(self.select_group_members(groups))

    if pkgs and not self.cpo.pargs.nodeps:
      deps, missing_deps = self.satisfy_dependencies(pkgs)
      pkgs |= deps
      if missing_deps:
        logging.error('failed to resolve the following dependencies: {}'.format(
          ' '.join(sorted(missing_deps))
        ))

    return XCPF.Pkg.PkgSet(pkg for pkg in pkgs if not pkg.is_ignored)


  def generate_build_script(self, pkgs):
    '''
    Generate build scripts. Insert the user-defined prologue then walk through
    the dependency-ordered list installing binary packages in batches with
    pacman and building packages using the user-defined build snippets. Clean
    up installed dependencies with a custom pacman clean command and then insert
    the epilogue if the user has defined one. Inserted Pacman scripts will use
    the "pacman" snippet if available so the user can configure pre- and post-
    pacman actions between snippets. This is useful for e.g. waiting for
    pending builds to complete before installing binary packages.
    '''
    snippets_conf = self.conf['build']['snippets']
    prologue_snippet_path = snippets_conf['prologue']
    epilogue_snippet_path = snippets_conf['epilogue']
    build_snippet_path = snippets_conf['build']
    pacman_snippet_path = snippets_conf['pacman']
    custom_build_snippet_paths = list(
      (re.compile(rgx), path) for rgx, path in snippets_conf['custom build']
    )
    pacman_cmd = self.conf['pacman']['command']
    sudo_pacman_cmd = self.conf['sudo']['command'] + pacman_cmd

    if pacman_snippet_path is None:
      pacman_snippet = None
    else:
      with open(pacman_snippet_path, 'r') as f:
        pacman_snippet = f.read()

    # The prologue path determines the extension. Presumably this script
    # starts with a hashbang and determines the execution.
    _, ext = os.path.splitext(prologue_snippet_path)
    script_path = os.path.join(self.pargs.pb_build_directory, 'build' + ext)

    # Map of paths to content.
    build_snippets = dict()

    logging.debug('copying prologue: {}'.format(prologue_snippet_path))
    os.makedirs(self.pargs.pb_build_directory, exist_ok=True)
    shutil.copy(prologue_snippet_path, script_path)

    # The databases have already been refreshed.
    self.cpo.pargs.refresh = 0

    cmd = None
    if self.pargs.pb_pre_install:
      # Determine binary command-line targets.
      bin_pkgs = set(
        itertools.chain.from_iterable(
          (pkg.qualified_pkgname(), pkg.pkgname())
          for pkg in pkgs if isinstance(pkg, XCPF.Pkg.PyalpmPkg)
        )
      )
      bin_pkg_args = set(arg for arg in self.cpo.pargs.args if arg in bin_pkgs)
      if bin_pkgs or self.cpo.pargs.sysupgrade:
        self.cpo.pargs.args = sorted(bin_pkg_args)
        cmd = sudo_pacman_cmd.copy()
        cmd.extend(XCPF.Pacman.unparse(self.cpo.pargs))

    # Remove flags from subsequent commands.
    self.cpo.pargs.sysupgrade = 0

    unrequired_deps = set()

    with open(script_path, 'a') as f:
      if cmd:
        f.write(self.interpolate_snippet(pacman_snippet, cmds=(cmd,)))

      # This loops over DependencyGraphNodes and lists of them.
      for x in self.structure_installation(pkgs):
        # Pacman sync packages. Pacman will pull in the dependencies
        if isinstance(x, list):
          # If the pre-install option has been passed then all binary packages
          # will be installed by a previous command.
          if self.pargs.pb_pre_install:
            continue
          deps = list(dgn.pkg.qualified_pkgname() for dgn in x if dgn.pkg.is_dep)
          if deps:
            unrequired_deps.update(dgn.pkg.pkgname() for dgn in x if dgn.pkg.is_dep)
            self.cpo.pargs.args = deps
            cmd = sudo_pacman_cmd.copy()
            cmd.append('--asdeps')
            cmd.extend(XCPF.Pacman.unparse(self.cpo.pargs))
            f.write(self.interpolate_snippet(pacman_snippet, cmds=(cmd,)))
          explicit = list(dgn.pkg.qualified_pkgname() for dgn in x if not dgn.pkg.is_dep)
          if explicit:
            self.cpo.pargs.args = explicit
            cmd = sudo_pacman_cmd.copy()
            cmd.extend(XCPF.Pacman.unparse(self.cpo.pargs))
            f.write(self.interpolate_snippet(pacman_snippet, cmds=(cmd,)))

        # Build packages.
        else:
          pkg = x.pkg
          if pkg.is_dep:
            unrequired_deps.add(pkg.pkgname())
          for rgx, path in custom_build_snippet_paths:
            if rgx.search(pkg.qualified_pkgname()):
              break
          else:
            path = build_snippet_path
          try:
            snippet = build_snippets[path]
          except KeyError:
            logging.debug('loading snippet: {}'.format(path))
            with open(path, 'r') as g:
              snippet = g.read()
            build_snippets[path] = snippet
          block = self.interpolate_snippet(snippet, dgn=x)
          f.write('\n')
          f.write(block)
          f.write('\n')

      # Clean up leftover unrequired deps which where not already installed.
      local_db = self.cpo.handle.get_localdb()
      unrequired_deps = sorted(
        shlex.quote(d) for d in unrequired_deps if not local_db.get_pkg(d)
      )

      if unrequired_deps:
        ls_unreq_cmd = pacman_cmd.copy()
        ls_unreq_cmd.extend(('-Qqd', '--config', self.cpo.pargs.config))
        ls_unreq_cmd.extend(unrequired_deps)

        rm_unreq_cmd = sudo_pacman_cmd.copy()
        rm_unreq_cmd.extend(('-Runs', '--config', self.cpo.pargs.config, '-'))

        cmd = '{} | {}'.format(
          XCGF.sh_quote_words(ls_unreq_cmd),
          XCGF.sh_quote_words(rm_unreq_cmd)
        )
        f.write(self.interpolate_snippet(pacman_snippet, cmds=(cmd,)))

      if epilogue_snippet_path:
        logging.debug('appending epilogue: {}'.format(epilogue_snippet_path))
        with open(epilogue_snippet_path, 'r') as e:
          epilogue = e.read()
        f.write(epilogue)

    # Set mode and owner.
    st = os.stat(script_path)
    os.chmod(script_path, st.st_mode | stat.S_IEXEC)
    uid, gid = self.get_build_user_uid_and_gid()
    if uid and gid:
      os.chown(script_path, uid, gid)
    return script_path




  def clean_script(self, remaining_deps):
    '''
    Create a script to remove leftover build dependencies.
    '''
    pacman_config = self.powerpill.pargs['pacman_config']
    sudo_cmd = get_bin(self.config, 'sudo')
    pacman_cmd = get_bin(self.config, 'pacman')
    rm_cmd = sudo_cmd + pacman_cmd + ['-Runs', '--config', pacman_config]
    lst_cmd = pacman_cmd + ['-Qq', '--config', pacman_config]
    return '''{}
target_pkgs_=(
  {}
)
installed_pkgs_=($({} "${{target_pkgs_[@]}}"))
{} "${{installed_pkgs_[@]}}"
'''.format(
      SCRIPT_HEADER,
      '\n  '.join(shlex.quote(rd) for rd in sorted(remaining_deps)),
      XCGF.sh_quote_words(lst_cmd),
      XCGF.sh_quote_words(rm_cmd)
    )



  def run(self):
    '''
    Perform the requested operation.

    Return True if successfull or False if not. This may pass control to Pacman.
    '''
    if self.pargs.pb_temporary_database:
      pacman_db = os.path.join(self.pargs.pb_temporary_database, 'pacman')
      other_db = os.path.join(self.pargs.pb_temporary_database, 'other')
      tsd = TemporarySyncDatabase(self, dbpath=pacman_db)
      self.env[XCPF.RepoWrapper.ENV_KEY_DATABASE_DIR] = other_db
      with tsd as pb:
        return pb._run()
    else:
      return self._run()


  def _run(self):
    '''
    Perform the requested operation.

    Return True if successfull or False if not. This may pass control to Pacman.
    '''
    if self.pargs.pb_clean:
      self.clean()
      return True

    if self.pargs.pb_export_default_config or self.pargs.pb_export_config:
      if self.pargs.pb_export_config:
        path = self.pargs.pb_export_config
        conf = self.conf
      else:
        path = self.pargs.pb_export_default_config
        conf = self.DEFAULT_CONF

      if path == '-':
        print(json.dumps(conf, indent='  ', sort_keys=True))
      else:
        with open(path, 'w') as f:
          json.dump(f, conf, indent='  ', sort_keys=True)
      return True

    if self.pargs.pb_snippet_help:
      self.print_snippet_interpolation_help()
      return True

    if not self.actionable():
      return self.run_pacman()

    if self.cpo.pargs.operation == 'sync':
      if self.cpo.pargs.refresh:
        self.refresh(force=(self.cpo.pargs.refresh>1))
        self.cpo.pargs.refresh = 0

      if self.cpo.pargs.search:
        self.display_search(self.cpo.pargs.args, by='name-desc', intersect=False)
        return True

      elif self.cpo.pargs.info:
        self.display_info(self.cpo.pargs.args, extended=self.cpo.pargs.info>1)
        return True

      elif self.cpo.pargs.list:
        self.display_list(repos=self.cpo.pargs.args)
        return True

      elif self.cpo.pargs.groups:
        self.display_groups(names=self.cpo.pargs.args, include_members=self.cpo.pargs.groups>1)
        return True

      else:
        pkgs = self.determine_sync_targets()

        if self.cpo.pargs.print:
          self.print(pkgs)
          return True

        elif not pkgs:
          return True

        if self.pargs.pb_download_pkgbuilds:
          return self.operation_download_pkgbuilds(pkgs)

        # Do not download binary packages but download PKGBUILDs.
        if self.pargs.pb_no_packages:
          download_pkgs, _ = self.partition_buildable(pkgs)
          self.download_packages_and_pkgbuilds(download_pkgs)
        else:
          found = self.download_packages_and_pkgbuilds(
            pkgs,
            download_construenda=self.pargs.pb_pre_install
          )
          not_found = pkgs - found
          if not_found:
            logging.error('not found: {}'.format(
              ' '.join(sorted(p.qualified_pkgname() for p in not_found))
            ))
            return False

        if self.cpo.pargs.downloadonly:
          return True

        if self.pargs.pb_gen_script or any(pkg.build for pkg in pkgs):
          script_path = self.generate_build_script(pkgs)
          logging.info('generated {}'.format(script_path))
          if self.pargs.pb_gen_script:
            return True
          else:
            return self.run_build_script(script_path)

    elif self.cpo.pargs.operation == 'query':
      if self.cpo.pargs.upgrades:
        self.display_upgradable(targets=self.cpo.pargs.args, downgradable=self.cpo.pargs.upgrades>1)
        return True

    return self.run_pacman()


  def run_pacman(self):
    '''
    Pass control to Pacman using the parsed Pacman arguments.
    '''
    self.stop_aria2_rpc_server()
    cmd = self.conf['pacman']['command'].copy()
    cmd.extend(XCPF.Pacman.unparse(self.cpo.pargs))
    try:
      subprocess.run(cmd)
    except subprocess.CalledProcessError:
      return False
    else:
      return True


  def get_build_user_uid_and_gid(self):
    '''
    Determine the user as which to run the build script, if there is one.
    '''
    user = self.conf['build']['user']
    if not user:
      user = os.getenv('SUDO_USER')
    if user:
      passwd = pwd.getpwnam(user)
      if passwd.pw_uid != os.getuid() \
      or passwd.pw_gid != os.getgid():
        return passwd.pw_uid, passwd.pw_gid
    return None, None


  def run_build_script(self, path):
    '''
    Run build script.
    '''
    # Drop root privileges if detected and possible.
    uid, gid = self.get_build_user_uid_and_gid()
    if gid:
      os.setgid(gid)
    if uid:
      os.setuid(uid)
    try:
      subprocess.run([path])
    except subprocess.CalledProcessError:
      return False
    else:
      return True



  def chown_for_builder(self, path):
    '''
    (Recursively) chown a path to the builder UID and GID.
    '''
    uid, gid = self.get_build_user_uid_and_gid()
    if not (uid and gid):
      return
    try:
      os.chown(path, uid, gid)
      st = os.stat(path)
    except FileNotFoundError:
      return
    if stat.S_ISDIR(st.st_mode):
      for root, dirs, files in os.walk(path):
        for item in itertools.chain(dirs, files):
          itempath = os.path.join(root, item)
          try:
            os.chown(itempath, uid, gid)
          except FileNotFoundError:
            pass



  def clean(self):
    '''
    Remove leftover control files.
    '''
    ID = self.get_id()
    pattern = '*{}'.format(Aria2JsonRpc.ARIA2_CONTROL_FILE_EXT)
    for locker in (
      self.cpo.db_path_and_lock,
      self.cpo.cache_path_and_lock,
    ):
      dirpath, lockpath = locker()
      try:
        with XCGF.Lockfile(lockpath, ID):
          for controlfile in glob.iglob(os.path.join(dirpath, pattern)):
            logging.info('removing {}'.format(controlfile))
            os.remove(controlfile)
      except XCGF.LockError as e:
        logging.warning('failed to clean {} [{}]'.format(dirpath, e))



##################################### Main #####################################

def main(args=None):
  parser = argparse.ArgumentParser(
    description='Extensible Pacman wrapper.',
    epilog='In addition to the above, all pacman options should be supported. See "pacman --help" for details. You may also prefix individual package name arguments with "{build}" and "{}" to indicate that the package should be built or not be built, respectively. E.g. "{build}core/linux" or just "{build}linux".'.format(
      Pacboy.NOBUILD_PREFIX,
      build=Pacboy.BUILD_PREFIX
    )
  )
  pb = Pacboy()
  pb.parse_args(parser)
  pb.configure_logging()
  with pb as x:
    return x.run()


if __name__ == '__main__':
  try:
    if not main():
      sys.exit('exited with error')
  except (KeyboardInterrupt, BrokenPipeError):
    pass
  except PacboyError as e:
    sys.exit(str(e))
