#!/usr/bin/env python3

import functools
import logging
import os
import pathlib
import shutil
import tarfile
import urllib.error
import urllib.parse

from XCGF import mirror
from XCPF.Pkg import PyalpmPkg, SrcinfoPkg
from XCPF.SRCINFO import Srcinfo
import XCPF.common
import XCPF.RepoWrapper

'''
A build extender for repos from http://xyne.archlinux.ca/repos/

The site provides SRCINFO files so we can use XCPF.Pkg.SrcinfoPkg instead of
subclassing XCPF.Pkg.Pkg ourselves.

This is currently implemented without caching. Ideally the SRCINFO should be
stored locally for the time specified by the ttl parameter.
'''



################################## CONSTANTS ###################################

SRCINFO_URL = 'http://xyne.archlinux.ca/projects/{}/pkgbuild/SRCINFO'
PKGBUILD_ARCHIVE_URL = 'http://xyne.archlinux.ca/projects/{}/pkgbuild.tar.gz'




################################### Classes ####################################

class XacBuildExtender(XCPF.RepoWrapper.BuildExtender):
  def __init__(self, env, repo):
    '''
    env:
      Same as RepoWrapperFactory.

    repo:
      Name of repo.
    '''
    super(XacBuildExtender, self).__init__(env)
    self.repo = repo


  # This should not be overridden.
  def cachepath(self, *subpath):
    '''
    Return the cache directory.
    '''
    path = XCPF.common.cachepath(str(self), *subpath)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return path


  # Override in a subclass.
  def srcinfo_url(self, pkgbase):
    '''
    Get the URL of the SRCINFO file.
    '''
    return SRCINFO_URL.format(urllib.parse.quote_plus(pkgbase))


  # Override in a subclass.
  def pkgbuild_archive_url(self, pkgbase):
    '''
    Get the URL of the PKGBUILD archive.
    '''
    return PKGBUILD_ARCHIVE_URL.format(urllib.parse.quote_plus(pkgbase))



  # Refresh metadata. Just delete the cached data.
  def refresh(self, force=False):
    dirpath = self.cachepath()
    try:
      shutil.rmtree(dirpath)
    except FileNotFoundError:
      pass
    # The refresh method may return download jobs. Yield nothing to ensure that
    # the function remains an iterator instead of returning None.
    yield from ()


  # Additional method not present in the base class.
  def get_srcinfo(self, pkgbase):
    '''
    Retrieve SRCINFO for the given pkgbase.
    '''
    # The URL of the SRCINFO file.
    url = self.srcinfo_url(pkgbase)

    # The path to the locally cached file.
    local_path = self.cachepath(pkgbase, os.path.basename(url))

    logging.info('mirroring {} to {}'.format(url, local_path))
    try:
      # Mirror the file to the cache.
      mirror(url, local_path, cache_time=self.env[XCPF.RepoWrapper.ENV_KEY_TTL])

    except urllib.error.HTTPError as e:
      if e.code != 304:
        logging.error(str(e))
        return None

    # Get a Srcinfo object that wraps SRCINFO.
    return Srcinfo(path=local_path)



  # Overridden method.
  def pkg(self, pkg):
    # Get the qualified pkgname.
    qualified_name = '{}/{}'.format(pkg.db.name, pkg.name)
    # Check if this package is a build target.
    if self.env[XCPF.RepoWrapper.ENV_KEY_IS_BUILD_TARGET](qualified_name):
      srcinfo = self.get_srcinfo(pkg.name)

      # Raise an error if the download failed.
      if not srcinfo:
        raise XCPF.RepoWrapper.RepoWrapperError(
          'failed to retrieve SRCINFO for {}'.format(qualified_name),
          self
        )

      # Get the architecture from the XCPF.Pacman.CommonPacmanOptions object.
      arch = self.env[XCPF.RepoWrapper.ENV_KEY_CPO].arch()

      # Extract the package metadata from the SRCINFO wrapper.
      pkginfo = srcinfo.pkginfo(pkg.name, arch=None)

      # Return a SrcinfoPkg based on this data.
      return SrcinfoPkg(True, self.repo, pkginfo)
    else:
      # Return a wrapped Pyalpm package if it's not a build target.
      return PyalpmPkg(pkg)



  def download_pkgbuilds(self, outdir, pkgbases, with_source_files=False):
    # This function is a little more complicated because it needs to generate
    # download jobs for parallel downloads.

    pkgbase_jobs = dict()

    if with_source_files:
      srcinfo_objs = list(self.get_srcinfo(pkgbase) for pkgbase in pkgbases)
    else:
      srcinfo_objs = None

    # Grouped downloads callback function.
    #
    # Each job associated with the callback will produce results. These results
    # are collected into a list. The first item in the list are the results from
    # the first job in the list of jobs, the second item are the results from
    # the second job, etc.
    #
    # The first job is for the PKGBUILD downloads. It is downloaded as an
    # archive and must be extracted. Any other jobs will be source downloads.
    #
    # The callback only checks if the PKGBUILD archive download has succeeded.
    # The source downloads can be dealt with by the user with makepkg later if
    # they fail.
    #
    # The callback must yield the qualified pkgbases that it was able to
    # download. Some jobs may download multiple pkgbases (e.g. via rsync). In
    # this case, it will always only be one, but for compatibility it still
    # needs to yield the result.
    def callback(pkgbase, results):
      # The first job returns a list of booleans with only 1 item here, because
      # there is only 1 job. See below.
      if all(results[0]):
        # The archive was downloaded successfully. Extract it to the target
        # path. Get the name from the url.
        url = self.pkgbuild_archive_url(pkgbase)
        archivepath = self.cachepath(pkgbase, os.path.basename(url))
        dirpath = os.path.join(outdir, pkgbase)
        with tarfile.open(archivepath) as t:
          # Strip the parent "pkgbuild" directory.
          for member in t.getmembers():
            current_path = pathlib.Path(member.name)
            stripped_path =  pathlib.Path(*current_path.parts[1:])
            member.name = str(stripped_path)
          t.extractall(dirpath)
        qualified_pkgbase = '{}/{}'.format(self.repo, pkgbase)
        logging.info('found {} at {}'.format(qualified_pkgbase, url))
        yield qualified_pkgbase


    for pkgbase in pkgbases:
      # Get the URL of the PKGBUILD archive.
      url = self.pkgbuild_archive_url(pkgbase)

      # Create a single download job for Aria2c.
      download = (
        'addUri',
        # A list of URLs. There is only one.
        [url],
        # Additional options.
        {
          # The name of the file.
          'out': os.path.basename(url),
          # The directory of the file.
          'dir': self.cachepath(pkgbase)
        }
      )

      # Each pkgbase job is a list of async jobs.
      # An async jobs is a 3-tuple of a method, its arguments and its keyword
      # arguments.
      #
      # self.env[XCPF.RepoWrapper.ENV_KEY_DOWNLOADER] is
      # Aria2JsonRpc.Aria2JsonRpc.polymethod_download_bool
      #
      # [[download]] is the list of arguments passed to that method. The first
      # argument is [download], i.e. a list of downloads to process in parallel.
      pkgbase_jobs[pkgbase] = [
        XCGF.AsyncJob(
          self.env[XCPF.RepoWrapper.ENV_KEY_DOWNLOADER],
          args=((download,),)
        )
      ]

    # Add source download jobs via the convencience method
    # append_source_download_jobs
    #
    # These will be either more Aria2c jobs or subprocess jobs that use the
    # VCS source downloader utility.
    if srcinfo_objs:
      XCPF.RepoWrapper.append_source_download_jobs(
        self.env, outdir, srcinfo_objs, pkgbase_jobs
      )

    # Create the list of grouped download jobs that will be passed to the
    # downloader.
    grouped_jobs = list(
      {
        'callback': functools.partial(callback, pkgbase),
        'jobs': jobs
      } for pkgbase, jobs in pkgbase_jobs.items()
    )

    # Empty set because existing PKGBUILDs are ignored. If they are still
    # "fresh" then the archives will just be re-extracted, so that is not an
    # issue. The group jobs will download what is needed.
    return set(), grouped_jobs




# The BuildExtenderFactor
class XacBuildExtenderFactory(XCPF.RepoWrapper.BuildExtenderFactory):
  def extend(self, env, reponame):
    return env, XacBuildExtender(env, reponame)
