#!/usr/bin/env python3
# -*- coding: utf8 -*-

# Copyright (C) 2012-2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from collections import OrderedDict
import argparse
import logging
import os
import platform
import pyalpm
import sys

import pycman.config
import XCGF
import XCPF.common


LIST_OPTIONS = pycman.config.LIST_OPTIONS + ('Usage',)


################################### Globals ####################################

COLOR_CHOICES = ('always', 'auto', 'never')
CUMULATIVE_LISTS = ('ignore', 'ignoregroup')

EXT_DB = '.db'
EXT_FILES_DB = '.files'
EXT_SIG = '.sig'
EXT_LOCK = '.lck'

DB_LOCK_FILE = 'db' + EXT_LOCK
CACHE_LOCK_FILE = 'cache' + EXT_LOCK

################################## Functions ###################################

def needs_sig(siglevel, prefix):
  '''
  Determine if a signature should be downloaded according to the SigLevel entry
  in pacman.conf.

  siglevel:
    The pacman.conf SigLevel entry.

  prefix:
    Either "Package" or "Database".

  The level is mapped to the following return types:

      * Never : False
      * Optional : None
      * Required : True

  '''
  if not siglevel:
    return None

  words = siglevel.split()
  pairs = (
    ('Never', False),
    ('Required', True),
    ('Optional', None)
  )
  # Check prefixed values first because they override non-prefixed values.
  for sl, rv in pairs:
    if (prefix + sl) in words:
      return rv
  for sl, rv in pairs:
    if sl in words:
      return rv
  return False



def split_commas(words):
  '''
  Split multi-argument fields from Pacman configuration file.
  '''
  for word in words:
    for w in word.split(','):
      if w:
        yield w



# TODO
# Use readline, curses or dialog to improve this.
def select_grp(grpname, pkgs, ignored=None):
  '''
  Present the user with a dialogue to select packages from a group.

  grpname: The name of the group.

  pkgs: A BuildablePkgSet.
  '''
  if not pkgs.pkgs:
    return pkgs
  selected = dict()
  for p in pkgs.pkgs:
    if ignored and p in ignored:
      selected[p] = False
    else:
      selected[p] = True
  names = sorted(selected)
  n = len(names)
  fmt = '[{}] {:' + str(len(str(n+1))) + 'd} {}'
  while True:
    print('Select packages from group "{}"'.format(grpname))
    for i in range(n):
      name = names[i]
      if selected[name]:
        marker = '*'
      else:
        marker = ' '
      print(fmt.format(marker, i+1, name))
    changes = input('Use +n to select and -n to deselect, where n is the package number.\nEnter nothing or "x" to finalize selection.\nMultiple entries are permitted.\nModify current selection: ')
    finalized = not changes
    for number in changes.split():
      if number == 'x':
        finalized = True
        continue
      try:
        number = int(number)
        index = abs(number) - 1
        if index >= n:
          raise ValueError
      except ValueError:
        logging.error('invalid number [{:d}]'.format(number))
        finalized = False
        break
      selected[names[index]] = (number > 0)
    print()
    if finalized:
      break

  for k, v in selected.items():
    if not v:
      del pkgs.pkgs[k]
  return pkgs



################################# PacmanConfig #################################
# This mess was originally a slight modification of the code in pyalpm to add
# support for signatures as it lagged behind Pacman updates. A patch was
# submitted upstream but after 2 years I never received a reply so this is here
# to stay for now.
#
# TODO
# Improve this, either by creating a C-binding to the ALPM parser or by
# refactoring it in clean Python.


def cb_log(level, line):
  '''
  Adaptation of pycman.config.cb_log to use the logging module.
  '''
  if level & pyalpm.LOG_ERROR:
    logging.error(line.strip())
  elif level & pyalpm.LOG_WARNING:
    logging.warning(line.strip())
  elif level & pyalpm.LOG_DEBUG or level & pyalpm.LOG_FUNCTION:
    logging.debug(line.strip())
#   elif level & pyalpm.LOG_FUNCTION:
#     logging.info(line.strip())



class PacmanConfig(OrderedDict):
  RELATIVE_PATHS = {
    'DBPath': 'var/lib/pacman',
    'LogFile': 'var/log/pacman.log'
  }


  def __init__(self, path=None, pargs=None):
    super(PacmanConfig, self).__init__()
    self.path = None
    self.pargs = None
    self.load(path=path, pargs=pargs)


  def load(self, path=None, pargs=None):
    self.reset()

    # This allows simply reloading with the given arguments.
    if pargs and pargs.config:
      path = pargs.config
    elif path:
      pass
    elif self.path:
      path = self.path
    else:
      path = XCPF.common.DEFAULT_PACMAN_CONF_PATH

    self.path = path
    if path is not None:
      self.load_from_file(path)
    # Arguments overwrite the configuration file.
    self.pargs = pargs
    if pargs is not None:
      self.load_from_pargs(pargs)


  def reset(self):
    self['options'] = OrderedDict()
    self.options = self['options']
    self.repos = OrderedDict()
#     self.options['RootDir'] = '/'
#     self.options['DBPath']  = '/var/lib/pacman'
    self.options['GPGDir']  = '/etc/pacman.d/gnupg/'
#     self.options['LogFile'] = '/var/log/pacman.log'
    self.options['IgnorePkg'] = list()
    self.options['IgnoreGroup'] = list()
    self.options['SigLevel'] = 'Optional'
#     self.options['Architecture'] = os.uname()[-1]
    self.options['Color'] = 'never'


  def get_option(self, option):
    '''
    Option getter that handles Pacman's path options as a function of which are
    set.
    '''
    if option == 'RootDir':
      return self.options.get(option, '/')
    elif option in self.RELATIVE_PATHS:
      try:
        return self.options[option]
      except KeyError:
        root = self.get_option('RootDir')
        return os.path.join(root, self.RELATIVE_PATHS[option])
    elif option == 'CacheDir':
      return self.options.get(option, ['/var/cache/pacman/pkg'])
    elif option == 'Architecture':
      arch = self.options.get(option, 'auto')
      return platform.machine() if arch == 'auto' else arch
    elif option in ('LocalFileSigLevel', 'RemoteFileSigLevel'):
      try:
        return self.options[option]
      except KeyError:
        return self.get_option('SigLevel')
    else:
      return self.options[option]


  def load_from_file(self, filename):
    for section, key, value in pycman.config.pacman_conf_enumerator(filename):
      if key == 'Architecture' and value == 'auto':
        continue
      elif key == 'Color':
        self[section][key] = 'auto'
        continue
      self.setdefault(section, OrderedDict())
      if key in LIST_OPTIONS:
        self[section].setdefault(key, []).append(value)
      else:
        self[section][key] = value
#     if 'CacheDir' not in self.options:
#       self.options['CacheDir']= ['/var/cache/pacman/pkg']
    for key, value in self.items():
      if key != 'options':
        if 'SigLevel' not in value:
          value['SigLevel'] = self.get_option('SigLevel')
        # For backwards compabilitility
        self.repos[key] = self[key]['Server']


  def load_from_pargs(self, pargs):
    for a, k in (
      ('root', 'RootDir'),
      ('dbpath', 'DBPath'),
      ('gpgdir', 'GPGDir'),
      ('arch', 'Architecture'),
      ('logfile', 'LogFile'),
      ('cachedir', 'CacheDir'),
      ('color', 'Color')
    ):
      v = getattr(pargs, a)
      if v is not None:
        self.options[k] = v
    if pargs.debug:
      pycman.config._logmask = 0xffff


  def apply(self, h):
    h.arch = self.get_option('Architecture')
    h.logfile = self.get_option('LogFile')
    h.gpgdir = self.get_option('GPGDir')
    h.cachedirs = self.get_option('CacheDir')
    h.logcb = cb_log

    # set sync databases
    for repo, servers in self.repos.items():
      db = h.register_syncdb(repo, 0)
      db_servers = []
      for rawurl in servers:
        url = rawurl.replace('$repo', repo)
        url = url.replace('$arch', self.get_option('Architecture'))
        db_servers.append(url)
      db.servers = db_servers

  def initialize_alpm(self):
    dbpath = self.get_option('DBPath')
    sync_db = os.path.join(dbpath, 'sync')
    local_db = os.path.join(dbpath, 'local')
    os.makedirs(sync_db, exist_ok=True)
    os.makedirs(local_db, exist_ok=True)
    h = pyalpm.Handle(self.get_option('RootDir'), dbpath)
    self.apply(h)
    return h

  def __str__(self):
    conf = ''
    for section, options in self.items():
      conf += '[%s]\n' % section
      for key, value in options.items():
        if key in LIST_OPTIONS:
          for v in value:
            conf += '%s = %s\n' % (key, v)
        elif key in pycman.config.BOOLEAN_OPTIONS:
          conf += key + '\n'
        else:
          conf += '%s = %s\n' % (key, value)
      conf += '\n'
    return conf



############################# CommonPacmanOptions ##############################

class CommonPacmanOptions(object):
  '''
  Wrapper for common Pacman configuration options and checks including runtime
  overrides such as target architectures set by command-line options.
  '''
  def __init__(self, pacman_conf=None):
    self.args = None
    self.pargs = None
    self.pacman_conf = pacman_conf
    self.handle = None


  def load_pacman_conf(self, path=None, pargs=None):
    '''
    Load a configuration object and pyalpm handle using the given arguments.
    '''
    if self.pacman_conf is None:
      self.pacman_conf = XCPF.Pacman.PacmanConfig(path=path, pargs=pargs)
    elif path:
      self.pacman_conf.load(path=path, pargs=pargs)
    elif pargs:
      self.pacman_conf.load(pargs=pargs)
    self.refresh()


  def load_pargs(self, pargs):
    '''
    Load parsed arguments.
    '''
    self.pargs = pargs
    self.load_pacman_conf(pargs=pargs)


  def parse_args(self, args=None, parent_parser=None):
    '''
    Load from command-line arguments.

    args:
      The arguments to pass to the parser.

    parent_parser:
      An argument parser to handle additional non-Pacman arguments.

    Returns the following:

        * Parsed parent arguments or None if parent_parser is None,
        * Parsed Pacman arguments if operation is either 'sync' nor 'query',
          otherwise a list of the remaining arguments with the "operation"
          attribute set to "other".
    '''
    if parent_parser:
      parent_pargs, args = parent_parser.parse_known_args(args)
    else:
      parent_pargs = None
      if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(add_help=False)
    parser = add_sync_args(parser)
    parser = add_common_args(parser)

    pargs, unknown = parser.parse_known_args(args)
    if pargs.operation == 'sync':
      if pargs.print_format:
        pargs.print = True
      if unknown:
        # Reparse to generate the appropriate error. If no error is generated
        # then these are just interspersed position arguments.
        parser.parse_args(unknown)
        pargs.args.extend(unknown)
    else:
      parser = argparse.ArgumentParser(add_help=False)
      parser = add_query_args(parser)
      parser = add_common_args(parser)
      pargs, unknown = parser.parse_known_args(args)
      if pargs.operation == 'query':
        if unknown:
          # Reparse to generate the appropriate error. If no error is generated
          # then these are just interspersed position arguments.
          parser.parse_args(unknown)
          pargs.args.extend(unknown)
      else:
        # Use this just to ensure that expected attributes exists.
        parser = argparse.ArgumentParser(add_help=False)
        parser = add_common_args(parser)
#         pargs = parser.parse_args([])
        pargs, _ = parser.parse_known_args(args)
    if pargs and pargs.args:
      pargs.args = list(XCPF.common.maybe_insert_args_from_stdin(pargs.args))
      pargs = format_lists(pargs)

    self.load_pargs(pargs)

    if pargs:
      self.pargs = pargs
    else:
      pargs = args

    return parent_pargs, pargs


  def refresh(self, force=False):
    '''
    Refresh metadata.
    '''
    self.handle = self.pacman_conf.initialize_alpm()


  def arch(self):
    '''
    Return the target architecture.
    '''
    return self.pacman_conf.get_option('Architecture')


  def ignored_pkgs(self):
    '''
    List ignored pkgnames.
    '''
    return self.pacman_conf.get_option('IgnorePkg')


  def ignored_groups(self):
    '''
    List ignored groups.
    '''
    return self.pacman_conf.get_option('IgnoreGroup')


  def is_ignored_pkg(self, name):
    '''
    Check if a package is ignored.
    '''
    return (name in self.ignored_pkgs())


  def is_ignored_group(self, name):
    '''
    Check if a group is ignored.
    '''
    return (name in self.ignored_groups())


  def is_ignored_pyalpm_pkg(self, pkg):
    '''
    Check if the given pyalpm package is ignored.
    '''
    return \
      self.is_ignored_pkg(pkg.name) or \
      any(is_ignoredgroup(g) for g in pkg.groups)


  def use_color(self):
    '''
    Determine if color should be used. Returns True or False.
    '''
    color = self.pacman_conf.get_option('Color')
    if color not in ('always', 'never', 'auto'):
      color = 'never'
    if color == 'never':
      return False
    elif color == 'always':
      return True
    else:
      return sys.stdout.isatty()


  def quiet(self):
    '''
    Return True if output should be quiet.
    '''
    return self.pacman_conf.pargs.quiet


  def usage(self, repo, operation):
    '''
    Check the pacman.conf Usage level for this repo and return a boolean
    indicating if the requested operation is permitted.
    '''
    try:
      u = self.pacman_conf[repo]['Usage']
    except KeyError:
      return True
    else:
      return (u and (operation in u or 'All' in u))


  def db_sigs(self, repo):
    '''
    Return True if the given repo expects database signatures.
    '''
    return needs_sig(self.pacman_conf[repo]['SigLevel'], 'Database')


  def pkg_sigs(self, repo):
    '''
    Return True if the given repo expects package signatures.
    '''
    return needs_sig(self.pacman_conf[repo]['SigLevel'], 'Package')


  def db_path_and_lock(self):
    path = self.pacman_conf.get_option('DBPath')
    return path, os.path.join(path, DB_LOCK_FILE)


  def cache_path_and_lock(self):
    for path in self.pacman_conf.get_option('CacheDir'):
      break
    return path, os.path.join(path, CACHE_LOCK_FILE)


  def expand_repos(self, repos):
    '''
    Expand lists of repos by replacing '*' with enabled Pacman repos. This may
    be extended in the future.

    Iterates over repo names.
    '''
    for r in repos:
      if r == '*':
        for db in self.handle.get_syncdbs():
          yield db.name
      else:
        yield r



############################### Common Arguments ###############################

def add_common_args(parser):
  parser.add_argument(
    'args', metavar='<arg>', nargs='*', default=[],
    help='arguments'
  )
  parser.add_argument(
    '-b', '--dbpath', metavar='<path>',
    help='set an alternate database location',
  )
  parser.add_argument(
    '-q', '--quiet', action='store_true',
    help='show less information for query and search',
  )
  parser.add_argument(
    '-r', '--root', metavar='<path>',
    help='set an alternate installation root',
  )
  parser.add_argument(
    '-v', '--verbose', action='store_true',
    help='be verbose',
  )
  parser.add_argument(
    '-a', '--arch', metavar='<arch>',
    help='set an alternate architecture',
  )
  parser.add_argument(
    '--cachedir', metavar='<path>', action='append',
    help='set an alternate package cache location',
  )
  parser.add_argument(
    '--color', choices=(COLOR_CHOICES),
    help='colorize the output',
  )
  parser.add_argument(
    '--config', metavar='<path>', default=XCPF.common.DEFAULT_PACMAN_CONF_PATH,
    help='set an alternate configuration file. Default: %(default)s',
  )
  parser.add_argument(
    '--confirm', action='store_true',
    help='always ask for confirmation',
  )
  parser.add_argument(
    '--debug', action='store_true',
    help='display debug messages',
  )
  parser.add_argument(
    '--gpgdir', metavar='<path>',
    help='set an alternate home directory for GnuPG',
  )
  parser.add_argument(
    '--hookdir', metavar='<dir>',
    help='set an alternate hook location',
  )
  parser.add_argument(
    '--logfile', metavar='<path>',
    help='set an alternate log file',
  )
  parser.add_argument(
    '--noconfirm', action='store_true',
    help='do not ask for any confirmation',
  )
  return parser



################################ Sync Arguments ################################

def add_sync_args(parser):
  parser.add_argument(
    '-S', '--sync', dest='operation', action='store_const', const='sync',
    help='pacman sync operation',
  )
  parser.add_argument(
    '-c', '--clean', action='count',
    help='remove old packages from cache directory (-cc for all)',
  )
  parser.add_argument(
    '-d', '--nodeps', action='count',
    help='skip dependency version checks (-dd to skip all checks)',
  )
  operation_group = parser.add_mutually_exclusive_group()
  operation_group.add_argument(
    '-g', '--groups', action='count',
    help='view all members of a package group (-gg to view all groups and members)',
  )
  operation_group.add_argument(
    '-i', '--info', action='count',
    help='view package information (-ii for extended information)',
  )
  operation_group.add_argument(
    '-l', '--list', action='store_true',
    help='view a list of packages in a repo',
  )
  # Print is not incompatible with the others operations.
  parser.add_argument(
    '-p', '--print', action='store_true',
    help='print the targets instead of performing the operation',
  )
  operation_group.add_argument(
    '-s', '--search', action='store_true',
    help='search remote repositories for matching strings',
  )
  operation_group.add_argument(
    '-u', '--sysupgrade', action='count',
    help='upgrade installed packages (-uu enables downgrades)',
  )
  parser.add_argument(
    '-w', '--downloadonly', action='store_true',
    help='download packages but do not install/upgrade anything',
  )
  parser.add_argument(
    '-y', '--refresh', action='count',
    help='download fresh package databases from the server (-yy to force a refresh even if up to date)',
  )
  reason_group = parser.add_mutually_exclusive_group()
  reason_group.add_argument(
    '--asdeps', action='store_true',
    help='install packages as non-explicitly installed',
  )
  reason_group.add_argument(
    '--asexplicit', action='store_true',
    help='install packages as explicitly installed',
  )
  parser.add_argument(
    '--assume-installed', metavar='<package=version>', action='append', default=[],
    help='add a virtual package to satisfy dependencies',
  )
  parser.add_argument(
    '--dbonly', action='store_true',
    help='only modify database entries, not package files',
  )
  parser.add_argument(
    '--force', action='store_true',
    help='force install, overwrite conflicting files',
  )
  parser.add_argument(
    '--ignore', metavar='<pkg>', action='append',
    help='ignore a package upgrade (can be used more than once)',
  )
  parser.add_argument(
    '--ignoregroup', metavar='<grp>', action='append',
    help='ignore a group upgrade (can be used more than once)',
  )
  parser.add_argument(
    '--needed', action='store_true',
    help='do not reinstall up to date packages',
  )
  parser.add_argument(
    '--noprogressbar', action='store_true',
    help='do not show a progress bar when downloading files',
  )
  parser.add_argument(
    '--noscriptlet', action='store_true',
    help='do not execute the install scriptlet if one exists',
  )
  parser.add_argument(
    '--print-format', metavar='<string>',
    help='specify how the targets should be printed',
  )
  # Undocumented
  parser.add_argument(
    '--ask', metavar='<bitmask>',
    help='undocumented option: control user prompts via a bitmask',
  )
  return parser



############################### Query Arguments ################################

# Various combinations of operations are incompatible but not in a way that
# can be represented by an argparse mutually exclusive group. For example, info
# and list are both incompatible with groups, but compatible with each other.
# Given that most query operations will be handled by Pacman, these conflicts
# may be ignored.
def add_query_args(parser):
  '''
  Pacman query options that depend on sync database data.
  '''
  parser.add_argument(
    '-Q', '--query', dest='operation', action='store_const', const='query',
    help='pacman query operation',
  )
  parser.add_argument(
    '-c', '--changelog', action='store_true',
    help='view the changelog of a package',
  )
  reason_group = parser.add_mutually_exclusive_group()
  reason_group.add_argument(
    '-d', '--deps', action='store_true',
    help='list packages installed as dependencies [filter]',
  )
  reason_group.add_argument(
    '-e', '--explicit', action='store_true',
    help='list packages explicitly installed [filter]',
  )
  parser.add_argument(
    '-g', '--groups', action='store_true',
    help='view all members of a package group',
  )
  parser.add_argument(
    '-i', '--info', action='count',
    help='view package information (-ii for backup files)',
  )
  parser.add_argument(
    '-k', '--check', action='count',
    help='check that package files exist (-kk for file properties)',
  )
  parser.add_argument(
    '-l', '--list', action='store_true',
    help='list the files owned by the queried package',
  )
  origin_group = parser.add_mutually_exclusive_group()
  origin_group.add_argument(
    '-m', '--foreign', action='store_true',
    help='list installed packages not found in sync db(s) [filter]',
  )
  origin_group.add_argument(
    '-n', '--native', action='store_true',
    help='list installed packages only found in sync db(s) [filter]',
  )
  parser.add_argument(
    '-o', '--owns', action='store_true',
    help='query the package that owns <file>',
  )
  parser.add_argument(
    '-p', '--file', action='store_true',
    help='query a package file instead of the database',
  )
  parser.add_argument(
    '-s', '--search', action='store_true',
    help='search locally-installed packages for matching strings',
  )
  parser.add_argument(
    '-t', '--unrequired', action='count',
    help='list packages not (optionally) required by any package (-tt to ignore optdepends) [filter]',
  )
  parser.add_argument(
    '-u', '--upgrades', action='store_true',
    help='list outdated packages [filter]',
  )
  return parser


############################# Argparser Functions ##############################

def flatten_list(xs):
  '''
  --ignore and --ignoregroup can be passed multiple times, and each time the
  argument may be a comma-separated list. This flattens the list.
  '''
  for x in xs:
    if ',' in x:
      for y in x.split(','):
        yield y.strip()
    else:
      yield x.strip()



def format_lists(pargs):
  '''
  Format cumulative lists by flattening them.
  '''
  for cl in CUMULATIVE_LISTS:
    try:
      xs = getattr(pargs, cl)
    except AttributeError:
      pass
    else:
      if xs:
        xs = list(flatten_list(xs))
        setattr(pargs, cl, xs)
  return pargs



def unparse(pargs, command=None):
  '''
  Rebuild an argument list from the parsed arguments.
  '''
  args = None
  for k, v in sorted(vars(pargs).items(), key=lambda x: x[0]):
    if k == 'args':
      args = v
    elif k == 'operation':
      yield '--' + v
    elif v is None or v is False:
      continue
    elif v is True:
      yield '--' + k
    elif isinstance(v, int):
      for i in range(v):
        yield '--' + k
    elif isinstance(v, list):
      for x in v:
        yield '--' + k
        yield x
    else:
      yield '--' + k
      yield v
  if args:
    yield from args
