#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2015-2016 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

'''
Common components of the XCPF module.
'''

import itertools
import logging
import os
import re
import sys
import textwrap
import types
import urllib.error
import urllib.request

import pyalpm
import xdg.BaseDirectory


import XCGF



################################## Constants ###################################

NAME = 'XCPF'

# Variant of ISO 8601 (2015-10-05T09:23:04.883Z)
ARCHLINUX_ORG_JSON_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


ARCHLINUX_ORG_JSON_URL = 'https://www.archlinux.org/packages/{repo}/{arch}/{pkgname}/json/'

ARCHLINUX_ORG_ABS_RSYNC_HOST = 'rsync.archlinux.org'
ARCHLINUX_ORG_ABS_RSYNC_PATH = '::abs/{arch}/{repo}/'

ARCHLINUX_ORG_ABS_GIT_HOST = 'https://projects.archlinux.org'
ARCHLINUX_ORG_ABS_GIT_TRUNK = '/svntogit/{gitrepo}.git/plain/{pkgbase}/trunk'

# Tuple elements:
# * repo name
# * repo architecture
# * is testing repo
# * Git repo name
ARCHLINUX_OFFICIAL_REPOS = (
  ('testing', 'any', True, 'packages'),
  ('core', 'any', False, 'packages'),
  ('extra', 'any', False, 'packages'),
  ('community-testing', 'any', True, 'community'),
  ('community', 'any', False, 'community'),
  ('multilib-testing', 'x86_64', True, 'community'),
  ('multilib', 'x86_64', False, 'community'),
)

ARCHLINUX_OFFICIAL_ARCHITECTURES = ('x86_64', 'i686')

ABS_GIT_REPOS = ('packages', 'community')

WRAP_WIDTH = 80

DEFAULT_PACMAN_CONF_PATH = '/etc/pacman.conf'
DEFAULT_TTL = 15 * 60


################################### Caching ####################################

def cachepath(*subdirs):
  '''
  Create a standard cache directory for this module and return a full path to
  the given subpath. Use this to re-use cached data across different scripts.
  '''
  cachedir = xdg.BaseDirectory.save_cache_path(NAME)
  return os.path.join(cachedir, *subdirs)



################################## Network ####################################

def archlinux_org_pkg_info_url(repo, arch, pkgname):
  '''
  Retrieve information about official packages from archlinux.org.
  '''
  return ARCHLINUX_ORG_JSON_URL.format(repo=repo, arch=arch, pkgname=pkgname)



def archlinux_org_pkg_info(repo, arch, pkgname):
  '''
  Retrieve information about official packages from archlinux.org.
  '''
  url = archlinux_org_pkg_info_url(repo, arch, pkgname)
  return XCGF.load_json_from_url(url)



def archlinux_abs_git_repos(repo=None):
  '''
  Iterate over the ABS git repos.
  '''
  if repo:
    for r, _, _, gr in ARCHLINUX_OFFICIAL_REPOS:
      if repo == r:
        yield gr
        break
  else:
    yield from ABS_GIT_REPOS



def search_abs_git_repos(pkgbase, repo=None):
  '''
  Scrape the online Git interface to retrieve URLs and names.
  '''
  for gr in archlinux_abs_git_repos(repo=repo):
    url = ARCHLINUX_ORG_ABS_GIT_HOST + ARCHLINUX_ORG_ABS_GIT_TRUNK.format(
      gitrepo=gr,
      pkgbase=pkgbase
    )
    logging.debug('querying {}'.format(url))
    try:
      with urllib.request.urlopen(url) as f:
        yield url, None
        for line in f:
          m = re.search(r"href='(.+?)'>(.+?)<".encode('utf-8'), line)
          if m:
            href = m.group(1).decode()
            name = m.group(2).decode()
            if name[:2] != '..':
              yield ARCHLINUX_ORG_ABS_GIT_HOST + href, name
    except urllib.error.HTTPError as e:
      if e.code != 404:
        raise e
    else:
      return



def retrieve_abs_via_git(output_dir, pkgbase, repo=None):
  '''
  Retrieve ABS files via the Git interface. Return the page url if successful.
  '''
  urls = search_abs_git_repos(pkgbase, repo=repo)
  try:
    url, _ = next(urls)
  except StopIteration:
    return None
  else:
    pkgdir = os.path.join(output_dir, pkgbase)
    os.makedirs(pkgdir, exist_ok=True)
    for href, name in urls:
      XCGF.mirror(href, os.path.join(pkgdir, name), cache_time=0)
    return url



def official_pyalpm_pkg_to_pkgbuild_url(pkg):
  '''
  Return a URL to the PKGBUILD of an official package.
  '''
  for repo, _, _, gitname in ARCHLINUX_OFFICIAL_REPOS:
    if repo == pkg.db.name:
      return ARCHLINUX_ORG_ABS_GIT_HOST + ARCHLINUX_ORG_ABS_GIT_TRUNK.format(
        gitrepo=gr,
        pkgname=pkg.name
      )
  return None



################################## Exceptions ##################################

class XcpfError(Exception):
  '''
  Parent class of all custom exceptions raised by this module.
  '''
  def __init__(self, msg, error=None):
    super(self.__class__, self).__init__()
    self.msg = msg
    self.error = error

  def __str__(self):
    s = '{}: {}'.format(self.__class__.__name__, self.msg)
    if self.error:
      s += ' [{}]'.format(self.error)
    return s



############################### pyalpm Database ################################

def join_db_pkgcaches(dbs):
  '''
  Create a unified list of packages from the given databases. The order is
  preserved.
  '''
  return list(itertools.chain.from_iterable(db.pkgcache for db in dbs))



def find_pkgname(dbs, pkgname):
  '''
  Search for a package in the given databases.
  '''
  for db in dbs:
    pkg = db.get_pkg(pkgname)
    if pkg is not None:
      return pkg
  return None


############################### Version Parsing ################################

def split_version_requirement(name):
  '''
  Split a name with a version requirement into a name, equality relation and
  version.
  '''
  for rel in ('>=', '<=', '>', '<', '='):
    try:
      name, ver = name.split(rel, 1)
      return name, rel, ver
    except ValueError:
      continue
  return name, None, None



def unsplit_version_requirement(name, rel, ver):
  '''
  Reverse of split_version_requirement()
  '''
  if rel is None:
    return name
  else:
    return '{}{}{}'.format(name, rel, ver)



def satisfies_version_requirement(version, relation, required_version):
  '''
  Check if a version requirement is satisfied by the given version.

  version:
    The version to check

  relation:
    The relation of the requirement, e.g. '>='.

  required_version:
    The required version.

  For example, satisfies_version_requirement('1.5', '>=', '1.4') will return
  True.
  '''
  if relation is None:
    return True
  else:
    vc = pyalpm.vercmp(version, required_version)
    return \
      (vc > 0 and '>' in relation) or \
      (vc == 0 and '=' in relation) or \
      (vc < 0 and '<' in relation)



def satisfied_by_names_and_versions(names_and_versions, req):
  '''
  Check if a requirement is satisfied by a set of package names and versions.
  This is useful for checking if virtually installed packages or packages in
  the installation queue satisfy the given requirement.

  names_and_versions:
    A map of package names (real or virtual) to versions.

  req:
    A possibly versioned dependency requirement.
  '''
  name, rel, ver = split_version_requirement(req)
  try:
    version = names_and_versions[name]
  except KeyError:
    return False
  else:
    return (version is None or satisfies_version_requirement(version, rel, ver))



def satisfies_all_version_requirements(version, reqs):
  '''
  Check if all version requirements are satisfied.
  '''
  return all(
    satisfies_version_requirement(version, *req) for req in reqs
  )



def collect_version_requirements(args):
  '''
  Collect version requirements per package.

  Returns a dictionary of package names mapped to lists of (relation, version)
  tuples for that name.
  '''
  ver_reqs = dict()

  for arg in args:
    name, relation, version = split_version_requirement(arg)
    if relation is None:
      if name not in ver_reqs:
        ver_reqs[name] = set()
    else:
      try:
        ver_reqs[name].add((relation, version))
      except KeyError:
        ver_reqs[name] = set(((relation, version),))

  return ver_reqs



def get_required_version_strings(ver_reqs, names):
  '''
  Return an iterator over the strings specifying the version requirements for
  the named package.
  '''
  for name in names:
    yielded = False
    try:
      for rel, ver in ver_reqs[name]:
        yield '{}{}{}'.format(name, rel, ver)
        yielded = True
    except KeyError:
      yield name
    else:
      if not yielded:
        yield name



def select_name_or_first(name, pkgs):
  '''
  Given a list of packages, return the package with the exact name if it exists,
  otherwise return the first package in the list.
  '''
  try:
    first = next(pkgs)
  except StopIteration:
    return None

  if first.name == name:
    return first

  for pkg in pkgs:
    if pkg.name == name:
      return pkg

  return first



def repo_and_package(pkgname):
  '''
  Split repo prefix from package name, e.g. "core/foo" -> "core", "foo".
  '''
  try:
    repo, name = pkgname.split('/', 1)
    return repo, name
  except ValueError:
    return None, pkgname



def find_satisfier(pkgs, req):
  '''
  Wrapper around pyalpm.find_satisfier that considers repo prefixes, e.g.
  core/foo.

  pkgs:
    Iterable of pyalpm packages.

  req:
    The requirement, e.g. bar/foo>=1.4

  Returns the satisfier or None.
  '''
  repo, name = repo_and_package(req)
  if repo is not None:
    pkgs = (p for p in pkgs if p.db.name == repo)
  return pyalpm.find_satisfier(pkgs, name)



def find_common_satisfiers_in_dbs(dbs, reqs, first_only=True):
  '''
  Search for packages that satisfy all of the given version requirements in the
  given databases.

  dbs:
    Iterable of pyalpm databases.

  reqs:
    Multiple requirements that must be satisfied simultaneously.

  first_only:
    If True, only return the first satisfier.

  Iterates over satisfiers.
  '''
  for db in dbs:
    yield from find_common_satisfiers(db.pkgcache, reqs, first_only=first_only)



def find_all_satisfiers(pkgs, req, already_stripped=False):
  '''
  pyalpm.find_satisfier only finds the first satisfier. This will find all.

  pkgs:
    An iterable of pyalpm packages to search for a satisfier.

  req:
    The requirement to satisfy, e.g. foo>=1.5

  already_stripped:
    If True then the repo has already been stripped from the qualified
    requirement.
  '''
  if not already_stripped:
    repo, req = repo_and_package(req)
    if repo is not None:
      pkgs = list(p for p in pkgs if p.db.name == repo)

  if not isinstance(pkgs, list):
    pkgs = list(pkgs)

  while pkgs:
    s = pyalpm.find_satisfier(pkgs, req)
    if s:
      yield s
      pkgs.remove(s)
    else:
      break



def find_common_satisfiers(pkgs, reqs, first_only=True):
  '''
  Search for packages that satisfy all of the given version requirements among
  the given packages.

  pkgs:
    Iterable of pyalpm packages.

  reqs:
    Iterable of requirements to satisfy simultaneously, e.g.
    ('foo>=1.0', 'foo<2.0')

  first_only:
    If True, yield only the first satisfier.

  Iterates over satisfiers, which may be None.
  '''
  if not (pkgs and reqs):
    return

  repos, reqs = zip(*(repo_and_package(r) for r in reqs))
  repos = set(r for r in repos if r)
  reqs = set(reqs)
  # If the requirements specify different repos then
  if len(repos) > 1:
    return

  # There will be at most 1.
  for r in repos:
    pkgs = list(p for p in pkgs if p.db.name == r)
    if not pkgs:
      return

  req = reqs.pop()
  if not reqs:
    for s in find_all_satisfiers(pkgs, req, already_stripped=True):
      yield s
      if first_only:
        return
    return

  for s in find_all_satisfiers(pkgs, req):
    ss = (s,)
    for r in reqs:
      if not pyalpm.find_satisfier(ss, r):
        break
    else:
      yield s
      if first_only:
        return






############################### Argument Parsing ###############################

def get_args_from_stdin():
  '''
  Get arguments from STDIN. This should emulate pacman's parsing of the "-"
  argument to insert package names piped via STDIN.
  '''
  for l in sys.stdin:
    yield from l.strip().split()



def maybe_insert_args_from_stdin(args):
  '''
  Replace "-" with arguments from STDIN if it appears in the given arguments.
  '''
  first = True
  for arg in args:
    if arg == '-' and first:
      first = False
      yield from get_args_from_stdin()
    else:
      yield arg



################################### Display ####################################

class InfoDisplay(object):
  '''
  Information formatter for package information. This emulates the format of
  pacman -Si and pacman -Qi.
  '''
  def __init__(self, width=WRAP_WIDTH, indent=0):
    self.width = width
    self.indent = indent
    self.wrapper = textwrap.TextWrapper(
      width=width,
      break_long_words=False,
      break_on_hyphens=False
    )

  def display(self, data):
    '''
    Print each line to STDOUT. Data should be an iterator over values of:

        <field name>, <field value>, <per line>

    where <field name> is the name of the package info field as a string, e.g.
    "Description", <field value> is either a string or an iterable, and
    <per line> is a boolean indicating if an each element of an iterable value
    should be displayed on a separate line.
    '''
    # The data needs to be traversed twice to determine the correct width of the
    # name column.

    if isinstance(data, types.GeneratorType):
      data = list(data)

    # Pacman inserts 2 spaces between the field name and the colon.
    name_width = max(len(x[0]) for x in data) + 2
    # Non-breaking space (U+00a0) should achieve the desired effect but textwrap
    # does not respect it. Instead, use an obscure Unicode character to enforce
    # the desired formatting. Without this, long words are shifted to the next
    # line but at the same position, e.g.
    #
    # Foobarwhatever :
    #                  http://example.com/some/really/long/path/to/something
    #
    # Use the word joiner character instead (U+2060)
    NBS = '\u2060' #'\u00a0'
    indent = NBS * self.indent
    spacer = ' ' * (self.indent + name_width + 2)
    self.wrapper.subsequent_indent = spacer

    def myprint(*args, **kwargs):
      print(*(a.replace(NBS, ' ') for a in args), **kwargs)

    for name, value, per_line in data:
      header = '{}{}:{}'.format(indent, name.ljust(name_width, NBS), NBS)

      if isinstance(value, str):
        pass
      elif not value:
        value = 'None'
        per_line = False
      else:
        try:
          value = sorted(value)
          if not per_line:
            value = '  '.join(value)
        except TypeError:
          value = str(value)
          per_line = False

      if per_line:
        for k, v in itertools.zip_longest((header,), value, fillvalue=spacer):
          if self.width <= 0:
            myprint(k, v, sep='')
          else:
            for l in self.wrapper.wrap(k + v):
              myprint(l)

      elif self.width <= 0:
        myprint(header, value, sep='')

      else:
        for l in self.wrapper.wrap(header + value):
          myprint(l)





def display_fields(pkg, extended=False):
  '''
  Iterate over package information display fields. This result must be accepted
  by InfoDisplay.display.

  "extended" is currently ignored because pyalpm is lagging behind ALPM and does
  not provide the additional extended information.
  '''
  yield 'Repository', pkg.db.name, False
  yield 'Name', pkg.name, False
  yield 'Version', pkg.version, False
  yield 'Description', pkg.desc, False
  yield 'Architecture', pkg.arch, False
  yield 'URL', pkg.url, False
  yield 'Licenses', pkg.licenses, False
  yield 'Groups', pkg.groups, False
  yield 'Provides', pkg.provides, False
  yield 'Depends On', pkg.depends, False
  yield 'Optional Deps', pkg.optdepends, True
  yield 'Conflicts With', pkg.conflicts, False
  yield 'Replaces', pkg.replaces, False
  yield 'Download Size', pkg.download_size, False
  yield 'Installed Size', pkg.isize, False
  yield 'Packager', pkg.packager, False
  yield 'Build Date', pkg.builddate, False

